
from .tab_pool import TabPooledChromium
from .chrome_context import ChromeContext

from .transport import ChromeExecutionManager
from .manager import ChromeRemoteDebugInterface
from .Generator import gen

from .cr_exceptions import ChromeControllerException
from .cr_exceptions import ChromeStartupException
from .cr_exceptions import ReusedPortError
from .cr_exceptions import ChromeConnectFailure
from .cr_exceptions import ChromeCommunicationsError
from .cr_exceptions import ChromeTabNotFoundError
from .cr_exceptions import ChromeError
from .cr_exceptions import ChromeDiedError
from .cr_exceptions import ChromeNavigateTimedOut
from .cr_exceptions import ChromeResponseNotReceived



import logging
import urllib.parse
import os
import contextlib

import cachetools
import threading


from ChromeController.manager import ChromeRemoteDebugInterface


class _TabStore(cachetools.LRUCache):
	def __init__(self, chrome_interface, *args, **kwargs):
		assert "maxsize" in kwargs
		assert kwargs['maxsize']
		super().__init__(*args, **kwargs)
		assert self.maxsize

		self.chrome_interface = chrome_interface
		self.log = logging.getLogger("Main.ChromeController.TabPool.Store")

	def __getitem__(self, key):
		self.log.debug("__getitem__: %s", key)
		assert key is not None, "You have to pass a key to __getitem__!"
		return super().__getitem__(key)

	def __missing__(self, key):
		self.log.debug("__missing__: %s", key)
		assert key is not None, "You have to pass a key to __missing__!"
		self[key] = (threading.Lock(), self.chrome_interface.new_tab())
		return self[key]

	def popitem(self):
		key, value = super().popitem()
		self.log.debug('Key "%s" evicted with value "%s"', key, value)
		dummy_lock, tab = value
		tab.close()
		return None


class TabPooledChromium(object):

	def __init__(self, *args, tab_pool_max_size = None, **kwargs):
		'''
		Create a chromium tab pool instance.

		This will start a chromium instance, from which new tabs will be created as
		needed with the tab() context manager.

		Note that the destruction of the `TabPooledChromium` object will kill the associated chromium
		execution. This will render any checked-out tabs invalid (though saving the tabs considering
		they're constructed in a context-manager is pretty obviously wrong anyways).
		'''
		if tab_pool_max_size is None:
			tab_pool_max_size = 10

		self.chrome_interface = ChromeRemoteDebugInterface(*args, **kwargs)
		self.tab_pool_max_size = tab_pool_max_size

		self.log = logging.getLogger("Main.ChromeController.TabPool")

		# We pass a tab to the tabstore, because otherwise it might wind up evicting the root tab,
		# which would take the entire chrome instance down with it when it's closed.
		self.__tab_cache = _TabStore(maxsize=tab_pool_max_size, chrome_interface=self.chrome_interface.new_tab())

		self.__counter_lock = threading.Lock()
		self.__active_tabs = {}

		self.__started_pid = os.getpid()

		self.alive = True

	def close(self):
		if self.alive:
			self.chrome_interface.close()
			self.alive = False

	def __del__(self):
		self.close()
		try:
			self.chrome_interface.close()
		except Exception:
			pass

	@contextlib.contextmanager
	def tab(self, netloc=None, url=None, extra_id=None, use_tid=False):
		'''
		Get a chromium tab from the pool, optionally one that has an association with a specific netloc/URL.

		If no url or netloc is specified, the per-thread identifier will be used.
		If `extra_id` is specified, it's stringified value will be mixed into the pool key
		If `use_tid` is true, the per-thread identifier will be mixed into the pool key.

		In all cases, the tab pool is a least-recently-used cache, so the tab that has been accessed the
		least recently will be automatically closed if a new tab is requested, and there are already
		`tab_pool_max_size` tabs created.

		'''
		assert self.alive, "Chrome has been shut down! Cannot continue!"
		if not netloc and url:
			netloc = urllib.parse.urlparse(url).netloc
			self.log.debug("Getting tab for netloc: %s (url: %s)", netloc, url)
		# Coerce to string type so even if it's none, it doesn't hurt anything.
		key = str(netloc)
		if extra_id:
			key += " " + str(extra_id)
		if use_tid or not key:
			key += " " + str(threading.get_ident())

		if self.__started_pid != os.getpid():
			self.log.error("TabPooledChromium instances are not safe to share across multiple processes.")
			self.log.error("Please create a new in each separate multiprocesssing process.")
			raise RuntimeError("TabPooledChromium instances are not safe to share across multiple processes.")

		with self.__counter_lock:
			self.__active_tabs.setdefault(key, 0)
			self.__active_tabs[key] += 1
			if self.__active_tabs[key] > 1:
				self.log.warning("Tab with key %s checked out more then once simultaneously", key)

		try:
			lock, tab = self.__tab_cache[key]
			with lock:
				yield tab
		finally:

			with self.__counter_lock:
				self.__active_tabs[key] -= 1
				if self.__active_tabs[key] == 0:
					self.__active_tabs.pop(key)




import logging
import sys
import click

from .Generator import gen
from . import chrome_context

@click.group()
@click.option('--verbose', '-v', is_flag=True, help='''
		The opposite of silent. Causes the internal logging to output \
		all traffic over the chromium control interface. VERY noisy.''')
@click.option('--silent', '-s', is_flag=True, help='''
		Suppress all output aside from the fetched content. \
		This can be used to make ChromeController act like \
		an alternative to curl with proper JS rendering.''')
def cli(verbose, silent):
	'''
ChromeController

\b
Usage: python3 -m ChromeController [-s | --silent] [-v | --verbose]
  python3 -m ChromeController fetch <url> [--binary <bin_name>] [--outfile <out_file_name>]
  python3 -m ChromeController update
  python3 -m ChromeController (-h | --help)
  python3 -m ChromeController --version

\b
Options:
  -s --silent   Suppress all output aside from the fetched content
                This basically makes ChromeController act like a alternative to curl
  -v --verbose  The opposite of silent. Causes the internal logging to output
                all traffic over the chromium control interface. VERY noisy.
  --version     Show version.
  fetch         Fetch a specified URL's content, and output it to the console.
	'''

	if verbose:
		logging.basicConfig(level=logging.DEBUG)
	elif silent:
		logging.basicConfig(level=logging.ERROR)
	else:
		logging.basicConfig(level=logging.INFO)


@cli.command()
def update():
	'''
	Update the generated class
	'''
	gen.update_generated_class(force=True, output_diff=True)

@cli.command()
def version():
	'''
	Print the ChromeController Version
	'''
	print("Version goes here?")

@cli.command()
@click.argument('url', required=True)
@click.option('--binary', '-b', default='google-chrome', help='Specify a specific chromium binary.')
@click.option('--outfile', '-o', default=False, help='''Write the output to a file. \
				Text content will be UTF-8 encoded. \
				Binary content will be written as-is.''')
@click.option('--noprint', '-n', is_flag=True, help='Do not print the fetched file to the console.')
@click.option('--rendered', '-r', is_flag=True, help='Allow the page to render in JS before extracting source.')
def fetch(url, binary, outfile, noprint, rendered):
	'''
	Fetch a specified URL's content, and output it to the console.
	'''
	with chrome_context.ChromeContext(binary=binary) as cr:
		resp = cr.blocking_navigate_and_get_source(url)
		if rendered:
			resp['content'] = cr.get_rendered_page_source()
			resp['binary'] = False
			resp['mimie'] = 'text/html'

	if not noprint:
		if resp['binary'] is False:
			print(resp['content'])
		else:
			print("Response is a binary file")
			print("Cannot print!")

	if outfile:
		with open(outfile, "wb") as fp:
			if resp['binary']:
				fp.write(resp['content'])
			else:
				fp.write(resp['content'].encode("UTF-8"))


if __name__ == '__main__':
	cli()






import distutils.spawn
import os.path
import sys
import subprocess
import pprint
import json
import base64
import signal
import pprint
import time
import http.cookiejar
import urllib.parse
import ChromeController.filter_funcs as filter_funcs

from ChromeController.cr_exceptions import ChromeResponseNotReceived
from ChromeController.cr_exceptions import ChromeNavigateTimedOut
from ChromeController.cr_exceptions import ChromeError
from ChromeController.resources import js


try:
	# Try to import the aparatus for generating the wrapper class, and
	# import it, if possible.
	from .Generator import gen
	gen.update_generated_class()
	ChromeRemoteDebugInterface_base = gen.get_class_def()
except ImportError:
	# If that failed, use the pre-generated version
	try:
		from ChromeController.Generator.Generated import ChromeRemoteDebugInterface as \
			ChromeRemoteDebugInterface_base
	except ImportError:
		raise RuntimeError("Generated class wrapper doesn't exist, and couldn't be created!")




DEFAULT_TIMEOUT_SECS = 10

class ChromeRemoteDebugInterface(ChromeRemoteDebugInterface_base):
	'''
	Remote control class for Chromium.
	'''

	def __init__(self,
		binary                = None,
		dbg_port              = None,
		use_execution_manager = None,
		visible_size          = None,
		disable_page          = False,
		disable_dom           = False,
		disable_network       = False,
		*args,
		**kwargs):
		super().__init__(
			binary                = binary,
			dbg_port              = dbg_port,
			use_execution_manager = use_execution_manager,
			*args, **kwargs)

		if disable_page:
			self.log.debug("Not enabling page debug interface")
		else:
			self.Page_enable()

		if disable_dom:
			self.log.debug("Not enabling DOM debug interface")
		else:
			self.DOM_enable()

		if disable_network:
			self.log.debug("Not enabling Network debug interface")
		else:
			self.Network_enable()

		if visible_size:
			assert isinstance(visible_size, tuple), "visible_size must be a 2-tuple containing 2 integers"
			assert len(visible_size) == 2, "visible_size must be a 2-tuple containing 2 integers"
			assert all([isinstance(val, int) for val in visible_size]), "visible_size must be a 2-tuple containing 2 integers"
			self.log.debug("Visible size overridden to %sx%s" % visible_size)
			self.Emulation_setVisibleSize(*visible_size)
		else:
			self.Emulation_setVisibleSize(1024, 1366)

		# cr_ver = self.Browser_getVersion()
		# self.log.debug("Remote browser version info:")
		# self.log.debug(str(cr_ver))
		# 'protocolVersion'
		# 'product'
		# 'revision'
		# 'userAgent'
		# 'jsVersion'


	def update_headers(self, header_args):
		'''
		Given a set of headers, update both the user-agent
		and additional headers for the remote browser.

		header_args must be a dict. Keys are the names of
		the corresponding HTTP header.

		return value is a 2-tuple of the results of the user-agent
		update, as well as the extra headers update.
		If no 'User-Agent' key is present in the new headers,
		the first item in the tuple will be None

		'''
		assert isinstance(header_args, dict), "header_args must be a dict, passed type was %s" \
			% (type(header_args), )

		ua = header_args.pop('User-Agent', None)
		ret_1 = None
		if ua:
			ret_1 = self.Network_setUserAgentOverride(userAgent=ua)


		ret_2 = self.Network_setExtraHTTPHeaders(headers = header_args)

		return (ret_1, ret_2)


	def __exec_js(self, script, args=None, **extra_params):
		'''

		Execute the passed javascript statement, optionally with passed
		arguments.

		Note that if `script` is not a function, it must be a single statement.
		The presence of semicolons not enclosed in a bracket scope will produce
		an error.

		'''

		if args is None:
			args = {}

		# How chromedriver does this:

		#  std::unique_ptr<base::Value>* result) {
		#   std::string json;
		#   base::JSONWriter::Write(args, &json);
		#   // TODO(zachconrad): Second null should be array of shadow host ids.
		#   std::string expression = base::StringPrintf(
		#       "(%s).apply(null, [null, %s, %s])",
		#       kCallFunctionScript,
		#       function.c_str(),
		#       json.c_str());

		expression = "({}).apply(null, [null, {}, {}])".format(
				js.kCallFunctionScript,
				script,
				json.dumps(args)
			)

		resp3 = self.Runtime_evaluate(expression=expression, **extra_params)

		return resp3




	# Interact with http.cookiejar.Cookie() instances
	def get_cookies(self):
		'''
		Retreive the cookies from the remote browser.

		Return value is a list of http.cookiejar.Cookie() instances.
		These can be directly used with the various http.cookiejar.XXXCookieJar
		cookie management classes.
		'''
		ret = self.Network_getAllCookies()

		assert 'result' in ret, "No return value in function response!"
		assert 'cookies' in ret['result'], "No 'cookies' key in function response"

		cookies = []
		for raw_cookie in ret['result']['cookies']:

			# Chromium seems to support the following key values for the cookie dict:
			# 	"name"
			# 	"value"
			# 	"domain"
			# 	"path"
			# 	"expires"
			# 	"httpOnly"
			# 	"session"
			# 	"secure"
			#
			#  This seems supported by the fact that the underlying chromium cookie implementation has
			#  the following members:
			#        std::string name_;
			#        std::string value_;
			#        std::string domain_;
			#        std::string path_;
			#        base::Time creation_date_;
			#        base::Time expiry_date_;
			#        base::Time last_access_date_;
			#        bool secure_;
			#        bool httponly_;
			#        CookieSameSite same_site_;
			#        CookiePriority priority_;
			#
			# See chromium/net/cookies/canonical_cookie.h for more.
			#
			# I suspect the python cookie implementation is derived exactly from the standard, while the
			# chromium implementation is more of a practically derived structure.

			# Network.setCookie

			baked_cookie = http.cookiejar.Cookie(
					# We assume V0 cookies, principally because I don't think I've /ever/ actually encountered a V1 cookie.
					# Chromium doesn't seem to specify it.
					version            = 0,

					name               = raw_cookie['name'],
					value              = raw_cookie['value'],
					port               = None,
					port_specified     = False,
					domain             = raw_cookie['domain'],
					domain_specified   = True,
					domain_initial_dot = False,
					path               = raw_cookie['path'],
					path_specified     = False,
					secure             = raw_cookie['secure'],
					expires            = raw_cookie['expires'],
					discard            = raw_cookie['session'],
					comment            = None,
					comment_url        = None,
					rest               = {"httponly":"%s" % raw_cookie['httpOnly']},
					rfc2109            = False
				)
			cookies.append(baked_cookie)

		return cookies

	def set_cookie(self, cookie):
		'''
		Add a cookie to the remote chromium instance.

		Passed value `cookie` must be an instance of `http.cookiejar.Cookie()`.
		'''

		# Function path: Network.setCookie
		# Domain: Network
		# Method name: setCookie
		# WARNING: This function is marked 'Experimental'!
		# Parameters:
		#         Required arguments:
		#                 'url' (type: string) -> The request-URI to associate with the setting of the cookie. This value can affect the default domain and path values of the created cookie.
		#                 'name' (type: string) -> The name of the cookie.
		#                 'value' (type: string) -> The value of the cookie.
		#         Optional arguments:
		#                 'domain' (type: string) -> If omitted, the cookie becomes a host-only cookie.
		#                 'path' (type: string) -> Defaults to the path portion of the url parameter.
		#                 'secure' (type: boolean) -> Defaults ot false.
		#                 'httpOnly' (type: boolean) -> Defaults to false.
		#                 'sameSite' (type: CookieSameSite) -> Defaults to browser default behavior.
		#                 'expirationDate' (type: Timestamp) -> If omitted, the cookie becomes a session cookie.
		# Returns:
		#         'success' (type: boolean) -> True if successfully set cookie.

		# Description: Sets a cookie with the given cookie data; may overwrite equivalent cookies if they exist.

		assert isinstance(cookie, http.cookiejar.Cookie), 'The value passed to `set_cookie` must be an instance of http.cookiejar.Cookie().' + \
			' Passed: %s ("%s").' % (type(cookie), cookie)

		# Yeah, the cookielib stores this attribute as a string, despite it containing a
		# boolean value. No idea why.
		is_http_only = str(cookie.get_nonstandard_attr('httponly', 'False')).lower() == "true"


		# I'm unclear what the "url" field is actually for. A cookie only needs the domain and
		# path component to be fully defined. Considering the API apparently allows the domain and
		# path parameters to be unset, I think it forms a partially redundant, with some
		# strange interactions with mode-changing between host-only and more general
		# cookies depending on what's set where.
		# Anyways, given we need a URL for the API to work properly, we produce a fake
		# host url by building it out of the relevant cookie properties.
		fake_url = urllib.parse.urlunsplit((
				"http" if is_http_only else "https",  # Scheme
				cookie.domain,                        # netloc
				cookie.path,                          # path
				'',                                   # query
				'',                                   # fragment
			))

		params = {
				'url'      : fake_url,

				'name'     : cookie.name,
				'value'    : cookie.value if cookie.value else "",
				'domain'   : cookie.domain,
				'path'     : cookie.path,
				'secure'   : cookie.secure,
				'expires'  : float(cookie.expires) if cookie.expires else float(2**32),

				'httpOnly' : is_http_only,

				# The "sameSite" flag appears to be a chromium-only extension for controlling
				# cookie sending in non-first-party contexts. See:
				# https://bugs.chromium.org/p/chromium/issues/detail?id=459154
				# Anyways, we just use the default here, whatever that is.
				# sameSite       = cookie.xxx
			}

		ret = self.Network_setCookie(**params)

		return ret

	def clear_cookies(self):
		'''
		At this point, this is just a thin shim around the Network_clearBrowserCookies() operation.

		That function postdates the clear_cookies() call here.
		'''
		self.Network_clearBrowserCookies()



	def navigate_to(self, url):
		'''
		Trigger a page navigation to url `url`.

		Note that this is done via javascript injection, and as such results in
		the `referer` header being sent with the url of the network location.

		This is useful when a page's navigation is stateful, or for simple
		cases of referrer spoofing.

		'''

		assert "'" not in url
		return self.__exec_js("window.location.href = '{}'".format(url))

	def get_current_url(self):
		'''
		Probe the remote session for the current window URL.

		This is primarily used to do things like unwrap redirects,
		or circumvent outbound url wrappers.

		'''
		res = self.Page_getNavigationHistory()
		assert 'result' in res
		assert 'currentIndex' in res['result']
		assert 'entries' in res['result']

		return res['result']['entries'][res['result']['currentIndex']]['url']

	def get_page_url_title(self):
		'''
		Get the title and current url from the remote session.

		Return is a 2-tuple: (page_title, page_url).

		'''

		cr_tab_id = self.transport._get_cr_tab_meta_for_key(self.tab_id)['id']
		targets = self.Target_getTargets()

		assert 'result' in targets
		assert 'targetInfos' in targets['result']

		for tgt in targets['result']['targetInfos']:
			if tgt['targetId'] == cr_tab_id:
				# {
				# 	'title': 'Page Title 1',
				# 	'targetId': '9d2c503c-e39e-42cc-b950-96db073918ee',
				# 	'attached': True,
				# 	'url': 'http://localhost:47181/with_title_1',
				# 	'type': 'page'
				# }

				title   = tgt['title']
				cur_url = tgt['url']
				return title, cur_url




	def click_link_containing_url(self, url):
		'''
		TODO

		'''

		# exec_func =

		self.__exec_js("window.location.href = '/test'")

		# js.kCallFunctionScript

		# "window.history.back();"

		# elem = self.find_element("//a".format(url))
		# print(elem)

	def execute_javascript(self, *args, **kwargs):
		'''
		Execute a javascript string in the context of the browser tab.
		'''

		ret = self.__exec_js(*args, **kwargs)
		return ret

	def find_element(self, search):

		'''
		DOM_performSearch(self, query, includeUserAgentShadowDOM)
		Python Function: DOM_performSearch
		        Domain: DOM
		        Method name: performSearch

		        WARNING: This function is marked 'Experimental'!

		        Parameters:
		                'query' (type: string) -> Plain text or query selector or XPath search query.
		                'includeUserAgentShadowDOM' (type: boolean) -> True to search in user agent shadow DOM.
		        Returns:
		                'searchId' (type: string) -> Unique search session identifier.
		                'resultCount' (type: integer) -> Number of search results.
		        Description: Searches for a given string in the DOM tree. Use <code>getSearchResults</code> to access search results or <code>cancelSearch</code> to end this search session.

		Python Function: DOM_getSearchResults
		        Domain: DOM
		        Method name: getSearchResults

		        WARNING: This function is marked 'Experimental'!

		        Parameters:
		                'searchId' (type: string) -> Unique search session identifier.
		                'fromIndex' (type: integer) -> Start index of the search result to be returned.
		                'toIndex' (type: integer) -> End index of the search result to be returned.
		        Returns:
		                'nodeIds' (type: array) -> Ids of the search result nodes.
		        Description: Returns search results from given <code>fromIndex</code> to given <code>toIndex</code> from the sarch with the given identifier.

		DOM_discardSearchResults(self, searchId)
		Python Function: DOM_discardSearchResults
		        Domain: DOM
		        Method name: discardSearchResults

		        WARNING: This function is marked 'Experimental'!

		        Parameters:
		                'searchId' (type: string) -> Unique search session identifier.
		        No return value.
		        Description: Discards search results from the session with the given id. <code>getSearchResults</code> should no longer be called for that search.
		'''

		res = self.DOM_performSearch(search, includeUserAgentShadowDOM=False)
		assert 'result' in res
		assert 'searchId' in res['result']
		searchid = res['result']['searchId']
		res_cnt  = res['result']['resultCount']
		self.log.debug("%s", res)
		self.log.debug("%s", searchid)

		if res_cnt == 0:
			return None

		items = self.DOM_getSearchResults(searchId=searchid, fromIndex=0, toIndex=res_cnt)

		self.log.debug("Results:")
		self.log.debug("%s", items)

		# DOM_getSearchResults


	def click_element(self, contains_url):
		'''

		TODO


		ChromeDriver source for how to click an element:

		Status ExecuteClickElement(Session* session,
		                           WebView* web_view,
		                           const std::string& element_id,
		                           const base::DictionaryValue& params,
		                           std::unique_ptr<base::Value>* value) {
		  std::string tag_name;
		  Status status = GetElementTagName(session, web_view, element_id, &tag_name);
		  if (status.IsError())
		    return status;
		  if (tag_name == "option") {
		    bool is_toggleable;
		    status = IsOptionElementTogglable(
		        session, web_view, element_id, &is_toggleable);
		    if (status.IsError())
		      return status;
		    if (is_toggleable)
		      return ToggleOptionElement(session, web_view, element_id);
		    else
		      return SetOptionElementSelected(session, web_view, element_id, true);
		  } else {
		    WebPoint location;
		    status = GetElementClickableLocation(
		        session, web_view, element_id, &location);
		    if (status.IsError())
		      return status;

		    std::list<MouseEvent> events;
		    events.push_back(
		        MouseEvent(kMovedMouseEventType, kNoneMouseButton,
		                   location.x, location.y, session->sticky_modifiers, 0));
		    events.push_back(
		        MouseEvent(kPressedMouseEventType, kLeftMouseButton,
		                   location.x, location.y, session->sticky_modifiers, 1));
		    events.push_back(
		        MouseEvent(kReleasedMouseEventType, kLeftMouseButton,
		                   location.x, location.y, session->sticky_modifiers, 1));
		    status =
		        web_view->DispatchMouseEvents(events, session->GetCurrentFrameId());
		    if (status.IsOk())
		      session->mouse_position = location;
		    return status;
		  }
		}
		'''

		pass

	def get_unpacked_response_body(self, requestId, mimetype="application/unknown"):
		'''
		Return a unpacked, decoded resposne body from Network_getResponseBody()
		'''
		content = self.Network_getResponseBody(requestId)

		assert 'result' in content
		result = content['result']

		assert 'base64Encoded' in result
		assert 'body' in result

		if result['base64Encoded']:
			content = base64.b64decode(result['body'])
		else:
			content = result['body']

		self.log.info("Navigate complete. Received %s byte response with type %s.", len(content), mimetype)

		return {'binary' : result['base64Encoded'],  'mimetype' : mimetype, 'content' : content}


	def handle_page_location_changed(self, timeout=None):
		'''
		If the chrome tab has internally redirected (generally because jerberscript), this
		will walk the page navigation responses and attempt to fetch the response body for
		the tab's latest location.
		'''

		# In general, this is often called after other mechanisms have confirmed
		# that the tab has already navigated. As such, we want to not wait a while
		# to discover something went wrong, so use a timeout that basically just
		# results in checking the available buffer, and nothing else.
		if not timeout:
			timeout = 0.1

		self.log.debug("We may have redirected. Checking.")

		messages = self.transport.recv_all_filtered(filter_funcs.capture_loading_events, tab_key=self.tab_id)
		if not messages:
			raise ChromeError("Couldn't track redirect! No idea what to do!")

		last_message = messages[-1]
		self.log.info("Probably a redirect! New content url: '%s'", last_message['params']['documentURL'])

		resp = self.transport.recv_filtered(filter_funcs.network_response_recieved_for_url(last_message['params']['documentURL'], last_message['params']['frameId']), tab_key=self.tab_id)
		resp = resp['params']

		ctype = 'application/unknown'

		resp_response = resp['response']

		if 'mimeType' in resp_response:
			ctype = resp_response['mimeType']
		if 'headers' in resp_response and 'content-type' in resp_response['headers']:
			ctype = resp_response['headers']['content-type'].split(";")[0]

		# We assume the last document request was the redirect.
		# This is /probably/ kind of a poor practice, but what the hell.
		# I have no idea what this would do if there are non-html documents (or if that can even happen.)
		return self.get_unpacked_response_body(last_message['params']['requestId'], mimetype=ctype)

	def blocking_navigate_and_get_source(self, url, timeout=DEFAULT_TIMEOUT_SECS):
		'''
		Do a blocking navigate to url `url`, and then extract the
		response body and return that.

		This effectively returns the *unrendered* page content that's sent over the wire. As such,
		if the page does any modification of the contained markup during rendering (via javascript), this
		function will not reflect the changes made by the javascript.

		The rendered page content can be retreived by calling `get_rendered_page_source()`.

		Due to the remote api structure, accessing the raw content after the content has been loaded
		is not possible, so any task requiring the raw content must be careful to request it
		before it actually navigates to said content.

		Return value is a dictionary with two keys:
		{
			'binary' : (boolean, true if content is binary, false if not)
			'content' : (string of bytestring, depending on whether `binary` is true or not)
		}

		'''


		resp = self.blocking_navigate(url, timeout)
		assert 'requestId' in resp
		assert 'response' in resp
		# self.log.debug('blocking_navigate Response %s', pprint.pformat(resp))

		ctype = 'application/unknown'

		resp_response = resp['response']

		if 'mimeType' in resp_response:
			ctype = resp_response['mimeType']
		if 'headers' in resp_response and 'content-type' in resp_response['headers']:
			ctype = resp_response['headers']['content-type'].split(";")[0]

		self.log.debug("Trying to get response body")
		try:
			ret = self.get_unpacked_response_body(resp['requestId'], mimetype=ctype)
		except ChromeError:
			ret = self.handle_page_location_changed(timeout)

		return ret


	def get_rendered_page_source(self, dom_idle_requirement_secs=3, max_wait_timeout=30):
		'''
		Get the HTML markup for the current page.

		This is done by looking up the root DOM node, and then requesting the outer HTML
		for that node ID.

		This calls return will reflect any modifications made by javascript to the
		page. For unmodified content, use `blocking_navigate_and_get_source()`

		dom_idle_requirement_secs specifies the period of time for which there must have been no
		DOM modifications before treating the rendered output as "final". This call will therefore block for
		at least dom_idle_requirement_secs seconds.
		'''

		# There are a bunch of events which generally indicate a page is still doing *things*.
		# I have some concern about how this will handle things like advertisements, which
		# basically load crap forever. That's why we have the max_wait_timeout.
		target_events = [
			"Page.frameResized",
			"Page.frameStartedLoading",
			"Page.frameNavigated",
			"Page.frameAttached",
			"Page.frameStoppedLoading",
			"Page.frameScheduledNavigation",
			"Page.domContentEventFired",
			"Page.frameClearedScheduledNavigation",
			"Page.loadEventFired",
			"DOM.documentUpdated",
			"DOM.childNodeInserted",
			"DOM.childNodeRemoved",
			"DOM.childNodeCountUpdated",
		]

		start_time = time.time()
		try:
			while 1:
				if time.time() - start_time > max_wait_timeout:
					self.log.debug("Page was not idle after waiting %s seconds. Giving up and extracting content now.", max_wait_timeout)
				self.transport.recv_filtered(filter_funcs.wait_for_methods(target_events),
					tab_key=self.tab_id, timeout=dom_idle_requirement_secs)

		except ChromeResponseNotReceived:
			# We timed out, the DOM is probably idle.
			pass



		# We have to find the DOM root node ID
		dom_attr = self.DOM_getDocument(depth=-1, pierce=False)
		assert 'result' in dom_attr
		assert 'root' in dom_attr['result']
		assert 'nodeId' in dom_attr['result']['root']

		# Now, we have the root node ID.
		root_node_id = dom_attr['result']['root']['nodeId']

		# Use that to get the HTML for the specified node
		response = self.DOM_getOuterHTML(nodeId=root_node_id)

		assert 'result' in response
		assert 'outerHTML' in response['result']
		return response['result']['outerHTML']


	def take_screeshot(self):
		'''
		Take a screenshot of the virtual viewport content.

		Return value is a png image as a bytestring.
		'''
		resp = self.Page_captureScreenshot()
		assert 'result' in resp
		assert 'data' in resp['result']
		imgdat = base64.b64decode(resp['result']['data'])
		return imgdat


	def blocking_navigate(self, url, timeout=DEFAULT_TIMEOUT_SECS):
		'''
		Do a blocking navigate to url `url`.

		This function triggers a navigation, and then waits for the browser
		to claim the page has finished loading.

		Roughly, this corresponds to the javascript `DOMContentLoaded` event,
		meaning the dom for the page is ready.


		Internals:

		A navigation command results in a sequence of events:

		 - Page.frameStartedLoading" (with frameid)
		 - Page.frameStoppedLoading" (with frameid)
		 - Page.loadEventFired" (not attached to an ID)

		Therefore, this call triggers a navigation option,
		and then waits for the expected set of response event messages.

		'''

		self.transport.flush(tab_key=self.tab_id)

		self.log.debug("Blocking navigate to URL: '%s'", url)
		ret = self.Page_navigate(url = url)

		assert("result"   in ret),           "Missing return content"
		assert("frameId"  in ret['result']), "Missing 'frameId' in return content"
		assert("loaderId" in ret['result']), "Missing 'loaderId' in return content"

		expected_id = ret['result']['frameId']
		loader_id   = ret['result']['loaderId']

		try:
			self.log.debug("Waiting for frame navigated command response.")
			self.transport.recv_filtered(filter_funcs.check_frame_navigated_command(expected_id), tab_key=self.tab_id, timeout=timeout)
			self.log.debug("Waiting for frameStartedLoading response.")
			self.transport.recv_filtered(filter_funcs.check_frame_load_command("Page.frameStartedLoading"), tab_key=self.tab_id, timeout=timeout)
			self.log.debug("Waiting for frameStoppedLoading response.")
			self.transport.recv_filtered(filter_funcs.check_frame_load_command("Page.frameStoppedLoading"), tab_key=self.tab_id, timeout=timeout)
			# self.transport.recv_filtered(check_load_event_fired, tab_key=self.tab_id, timeout=timeout)

			self.log.debug("Waiting for responseReceived response.")
			resp = self.transport.recv_filtered(filter_funcs.network_response_recieved_for_url(url=None, expected_id=expected_id), tab_key=self.tab_id, timeout=timeout)

			if resp is None:
				raise ChromeNavigateTimedOut("Blocking navigate timed out!")

			return resp['params']
		# The `Page.frameNavigated ` event does not get fired for non-markup responses.
		# Therefore, if we timeout on waiting for that, check to see if we received a binary response.
		except ChromeResponseNotReceived:
			# So this is basically broken, fix is https://bugs.chromium.org/p/chromium/issues/detail?id=831887
			# but that bug report isn't fixed yet.
			# Siiiigh.
			self.log.warning("Failed to receive expected response to navigate command. Checking if response is a binary object.")
			resp = self.transport.recv_filtered(
				keycheck = filter_funcs.check_frame_loader_command(
						method_name = "Network.responseReceived",
						loader_id   = loader_id
					),
				tab_key  = self.tab_id,
				timeout  = timeout)

			return resp['params']






import unittest
import socket
import json
import base64
import zlib
import gzip
import ChromeController
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread

from . import testing_server


CHROME_BINARY_NAME = "google-chrome"

class TestChromium(unittest.TestCase):
	# def setUp(self):
	# 	self.cr = ChromeController.TabPooledChromium("google-chrome")

	# def tearDown(self):
	# 	del self.cr

	def fetch_check_headers(self, expect_headers):

		try:
			# Configure mock server.
			self.mock_server_port, self.mock_server, self.mock_server_thread = testing_server.start_server(self, expect_headers)
			tgturl = "http://localhost:{}".format(self.mock_server_port)
			with ChromeController.ChromeContext(CHROME_BINARY_NAME) as cr:
				cr.update_headers(expect_headers)
				resp = cr.blocking_navigate_and_get_source(tgturl)

			self.assertEqual(resp['content'], 'Root OK?')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/html")
		finally:
			self.mock_server.shutdown()


	def test_basic_fetch_1(self):
		self.fetch_check_headers({})


	def test_custom_ua_1(self):
		'''
		Dumb basic check
		'''
		expect_headers = {
			'User-Agent' : r"Test test testy test testttttttt"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_ua_2(self):
		'''
		Special chars to see if we can intentionally break something
		'''
		expect_headers = {
			'User-Agent' : r"Test !@#$%^&*(;;);_;+;\\///\\ \"':>?<|}{][;;][[p\\tblah"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_ua_3(self):
		'''
		What if it's empty?
		'''
		expect_headers = {
			'User-Agent' : r""
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_ua_4(self):
		'''
		Or ridiculously long
		'''
		expect_headers = {
			'User-Agent' : r"wat" * 5000
		}
		self.fetch_check_headers(expect_headers)


	def test_custom_ua_5(self):
		'''
		Or something that looks like a accept header
		'''
		expect_headers = {
			'User-Agent' : r"text/html, application/xhtml+xml, application/xml;q=0.9, */*, */*, */*, */*, */*, */*, */*, */*, */* "
		}
		self.fetch_check_headers(expect_headers)


	def test_custom_lang_1(self):
		'''
		Normal lang
		'''
		expect_headers = {
			'Accept-Language' : r"en-US,en;q=0.9"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_lang_2(self):
		'''
		Special chars
		'''
		expect_headers = {
			'Accept-Language' : r"Test !@#$%^&*(;;);_;+;\\///\\ \"':>?<|}{][;;][[p\\tb ''' \"\" \" lah"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_lang_3(self):
		'''
		What if it's empty?
		'''
		expect_headers = {
			'Accept-Language' : r""
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_lang_4(self):
		'''
		Or ridiculously long
		'''
		expect_headers = {
			'Accept-Language' : r"wat" * 5000
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_lang_5(self):
		'''
		Or something that looks like a accept header
		'''
		expect_headers = {
			'Accept-Language' : r"text/html, application/xhtml+xml, application/xml;q=0.9, */*, */*, */*, */*, */*, */*, */*, */*, */* "
		}
		self.fetch_check_headers(expect_headers)


	def test_custom_accept_1_1(self):
		'''
		Normal accept
		'''
		expect_headers = {
			'Accept-Encoding' : r"text/html, application/xhtml+xml, application/xml;q=0.9, */*;q=0.8"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_1_2(self):
		expect_headers = {
			'Accept-Encoding' : r"text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_1_3(self):
		expect_headers = {
			'Accept-Encoding' : r"text/html,application/xhtml+xml, application/xml;q=0.9, */*;q=0.8"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_2_1(self):
		'''
		Normal accept
		'''
		expect_headers = {
			'Accept' : r"text/html, application/xml;q=0.9, application/xhtml+xml, image/png, image/webp, image/jpeg, image/gif, image/x-xbitmap, */*;q=0.8"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_2_2(self):
		expect_headers = {
			'Accept' : r"text/html, application/xml;q=0.9,application/xhtml+xml,image/png,image/webp,image/jpeg, image/gif, image/x-xbitmap, */*;q=0.8"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_2_3(self):
		expect_headers = {
			'Accept' : r"text/html, application/xml;q=0.9,application/xhtml+xml, image/png,image/webp,image/jpeg, image/gif, image/x-xbitmap,   */*;q=0.8"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_3(self):
		'''
		Send garbage
		'''
		expect_headers = {
			'Accept' : r"Test !@#$%^&*(;;);_;+;\\///\\ \"':>?<|}{][;;][[p\\tb ''' \"\" \" lah"
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_4(self):
		'''
		What if it's empty?
		'''
		expect_headers = {
			'Accept' : r""
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_5(self):
		'''
		Or ridiculously long/repeated
		'''
		expect_headers = {
			'Accept' : r"text/html, " * 5
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_accept_6(self):
		'''
		I was sending this as a bug at one point
		'''
		expect_headers = {
			'Accept' : r"text/html, application/xhtml+xml, application/xml;q=0.9, */*, */*, */*, */*, */*, */*, */*, */*, */* "
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_1_1(self):
		'''
		Normal accept
		'''
		expect_headers = {
			'Accept-Encoding' : r'gzip'
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_1_2(self):
		expect_headers = {
			'Accept-Encoding' : r'gzip, deflate'
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_1_3(self):
		expect_headers = {
			'Accept-Encoding' : r'deflate, gzip'
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_1_4(self):
		expect_headers = {
			'Accept-Encoding' : r'gzip,    deflate'
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_1_5(self):
		expect_headers = {
			'Accept-Encoding' : r'gzip, deflate, sdch'
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_1_6(self):
		expect_headers = {
			'Accept-Encoding' : r'gzip,deflate, sdch'
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_1_7(self):
		expect_headers = {
			'Accept-Encoding' : r'sdch, gzip,deflate'
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_2(self):
		'''
		Send the wrong header
		'''
		expect_headers = {
			'Accept-Encoding' : r"text/html, application/xml;q=0.9, application/xhtml+xml, image/png, image/webp, image/jpeg, image/gif, image/x-xbitmap, */*;q=0.8"
		}
		self.fetch_check_headers(expect_headers)

	# def test_custom_encoding_3(self):
	# 	'''
	# 	Send garbage
	# 	'''
	# 	expect_headers = {
	# 		'Accept-Encoding' : r"Test !@#$%^&*(;;);_;+;\\///\\ \"':>?<|}{][;;][[p\\tb ''' \"\" \" lah"
	# 	}
	# 	self.fetch_check_headers(expect_headers)

	def test_custom_encoding_4(self):
		'''
		What if it's empty?
		'''
		expect_headers = {
			'Accept-Encoding' : r""
		}
		self.fetch_check_headers(expect_headers)

	def test_custom_encoding_5(self):
		'''
		Or ridiculously long/repeated
		'''
		expect_headers = {
			'Accept-Encoding' : r"gzip,deflate, sdch, " * 5
		}
		self.fetch_check_headers(expect_headers)


	def test_setting_referrer_1(self):
		'''
		Referrers. See https://bugs.chromium.org/p/chromium/issues/detail?id=795336 ?
		'''
		expect_headers = {
			'Referer' : r"http://www.googlez.com"
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_referrer_2(self):
		expect_headers = {
			'Referer' : r"htt;ljksdfhglkjshdg!@#$%^&*()_++_)(*&^%$#@!}{\":>?><|{|}{\\][\';//.,1209-82409587p://www.googlez.com"
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_referrer_3(self):
		expect_headers = {
			'Referer' : r"http://www.googlez.com"*2
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_referrer_4(self):
		expect_headers = {
			'Referer' : r""
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_host_1(self):
		expect_headers = {
			'Host' : r""
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_host_2(self):
		expect_headers = {
			'Host' : r"http://www.googlez.com"
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_host_3(self):
		expect_headers = {
			'Host' : r"www.googlez.com"
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_host_4(self):
		expect_headers = {
			'Host' : r"htt;ljksdfhglkjshdg!@#$%^&*()_++_)(*&^%$#@!}{\":>?><|{|}{\\][\';//.,1209-82409587p://www.googlez.com"
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_host_5(self):
		expect_headers = {
			'Host' : r"www.googlez.com"*50
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_random_1(self):
		expect_headers = {
			'Pineapple' : r"Banana"
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_random_2(self):
		expect_headers = {
			'Pineapple'*5 : r"Banana"*5
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_random_3(self):
		expect_headers = {
			'Pineapple'*500 : r"Banana"*500
		}
		self.fetch_check_headers(expect_headers)

	def test_setting_random_4(self):
		expect_headers = {
			'Pineapple' : r"htt;ljksdfhglkjshdg!@#$%^&*()_++_)(*&^%$#@!}{\":>?><|{|}{\\][\';//.,1209-82409587p://www.googlez.com"
		}
		self.fetch_check_headers(expect_headers)


	# def test_setting_random_5(self):
	# 	expect_headers = {
	# 		'Pineapple;Pineapple' : r"htt;ljksdfhglkjshdg!@#$%^&*()_++_)(*&^%$#@!}{\":>?><|{|}{\\][\';//.,1209-82409587p://www.googlez.com"
	# 	}
	# 	self.fetch_check_headers(expect_headers)

	# def test_setting_random_6(self):
	# 	expect_headers = {
	# 		'Pineapple=Pineapple' : r"htt;ljksdfhglkjshdg!@#$%^&*()_++_)(*&^%$#@!}{\":>?><|{|}{\\][\';//.,1209-82409587p://www.googlez.com"
	# 	}
	# 	self.fetch_check_headers(expect_headers)

	# def test_setting_random_7(self):
	# 	expect_headers = {
	# 		'Pineapple=Pineapplehtt;ljksdfhglkjshdg!@#$%^&*()_++_)(*&^%$#@!}{\":>?><|{|}{\\][\';//.,1209-82409587p://www.googlez.com'
	# 			: r"htt;ljksdfhglkjshdg!@#$%^&*()_++_)(*&^%$#@!}{\":>?><|{|}{\\][\';//.,1209-82409587p://www.googlez.com"
	# 	}
	# 	self.fetch_check_headers(expect_headers)



import unittest
import socket
import json
import base64
import zlib
import gzip
import ChromeController
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread

from . import testing_server


CHROME_BINARY_NAME = "google-chrome"

class TestChromium(unittest.TestCase):
	def setUp(self):
		self.cr = ChromeController.TabPooledChromium("google-chrome")
		self.mock_server_port, self.mock_server, self.mock_server_thread = testing_server.start_server(self, {})

	def tearDown(self):
		self.mock_server.shutdown()
		del self.cr

	def test_basic_fetch_1(self):

		# Configure mock server.
		tgturl = "http://localhost:{}".format(self.mock_server_port)
		with ChromeController.ChromeContext(CHROME_BINARY_NAME) as cr:
			cr.update_headers({})
			resp = cr.blocking_navigate_and_get_source(tgturl)

		self.assertEqual(resp['content'], 'Root OK?')
		self.assertEqual(resp['binary'], False)
		self.assertEqual(resp['mimetype'], "text/html")


	def test_tab_repeatability_1(self):
		tgturl = "http://localhost:{}/".format(self.mock_server_port)
		with self.cr.tab(url=tgturl) as tab:
			resp = tab.blocking_navigate_and_get_source(tgturl)

		self.assertEqual(resp['content'], 'Root OK?')
		self.assertEqual(resp['binary'], False)
		self.assertEqual(resp['mimetype'], "text/html")

		print("Creating tab again!")
		with self.cr.tab(url=tgturl) as tab:
			at_url = tab.get_current_url()
			self.assertEqual(at_url, tgturl)


		print("3rd tab context!")
		with self.cr.tab(url=tgturl) as tab:
			title, cur_url = tab.get_page_url_title()
			print("title, cur_url", title, cur_url)
			self.assertEqual(cur_url, tgturl)


	def test_tab_flushing_1(self):
		tgturl = "http://localhost:{}/".format(self.mock_server_port)
		with self.cr.tab(url=tgturl) as tab:
			resp = tab.blocking_navigate_and_get_source(tgturl)

		self.assertEqual(resp['content'], 'Root OK?')
		self.assertEqual(resp['binary'], False)
		self.assertEqual(resp['mimetype'], "text/html")

		for x in range(20):
			print("Creating tab again!")
			with self.cr.tab(url=tgturl, extra_id=x) as tab:
				title, cur_url = tab.get_page_url_title()
				print("title, cur_url", title, cur_url)



		print("3rd tab context!")
		with self.cr.tab(url=tgturl) as tab:
			title, cur_url = tab.get_page_url_title()
			print("title, cur_url", title, cur_url)
			self.assertNotEqual(cur_url, tgturl)

	def test_tab_flushing_2(self):
		tgturl = "http://localhost:{}/".format(self.mock_server_port)
		with self.cr.tab(url=tgturl) as tab:
			resp = tab.blocking_navigate_and_get_source(tgturl)

		self.assertEqual(resp['content'], 'Root OK?')
		self.assertEqual(resp['binary'], False)
		self.assertEqual(resp['mimetype'], "text/html")

		for x in range(100):
			print("Creating tab again!")
			with self.cr.tab(url=tgturl, extra_id=x) as tab:
				title, cur_url = tab.get_page_url_title()
				print("title, cur_url", title, cur_url)



		# At this point, we should have fewer then 100 tabs, if nothing is broken.
		print("3rd tab context!")
		with self.cr.tab(url=tgturl) as tab:
			targets = tab.Target_getTargets()
			assert 'result' in targets
			assert 'targetInfos' in targets['result']

			self.assertLess(len(targets['result']['targetInfos']), 20)




import time
import traceback
import pprint
import gc
import uuid
import logging

from . import cr_exceptions
from .transport import ChromeExecutionManager



class ChromeInterface():
	"""
	Document me, maybe?
	"""


	def __init__(self, binary, dbg_port, use_execution_manager, *args, **kwargs):
		"""
		Base chromium transport initialization.

		The binary to execute is assumed to be named `chromium`, and on $PATH
		if not specified in the `binary` parameter.

		The chromium binary is launched with the arg `--remote-debugging-port={dbg_port}` if found.

		Note that the dbg_port must be GLOBALLY unique on a PER-COMPUTER basis. If not specified, it
		will default to an unused port >= 9222.

		Duplication of the dbg_port parameter can often lead to cr_exceptions.ChromeStartupException
		exceptions. If these happen, you may need to call ChromeInterface.close() to force shutdown
		of chromium instances, if you are not trying to instantiate multiple instances of chromium
		at once.

		"""

		# Force cleanup of dangling processes (if any)
		# This is needed because the deletion of active chromium processes
		# can occur dignificantly after the corresponding object goes out-of-scope, which
		# leads to connecting to the wrong process, and then when it is GC'ed,
		# you get a connection-died error.
		gc.collect()


		self.log = logging.getLogger("Main.ChromeController.Interface")
		if use_execution_manager:
			self.is_root_session = False
			self.transport, self.tab_id = use_execution_manager
			# self.transport.connect(tab_key = self.tab_id)

		else:
			self.is_root_session = True
			self.log.debug("Binary: %s", binary)
			self.log.debug("Args: %s", args)
			self.log.debug("Kwargs: %s", kwargs)

			self.tab_id = uuid.uuid4()

			self.transport = ChromeExecutionManager(
					binary       = binary,
					port         = dbg_port,
					base_tab_key = self.tab_id,
					*args,
					**kwargs
				)
			self.transport.check_process_ded()


	def __check_ret(self, ret):
		if ret is False or ret is None:
			raise cr_exceptions.ChromeError("Null response from Chromium (or timed out)!")

		if 'error' in ret:
			err = pprint.pformat(ret)
			raise cr_exceptions.ChromeError("Error in response: \n{}".format(err))

		self.log.debug("No exception in command response!")


	def synchronous_command(self, *args, **kwargs):
		'''
		Forward a command to the remote chrome instance via the transport
		connection, and check the return for an error.

		If the command resulted in an error, a `ChromeController.ChromeError` is raised,
		with the error string containing the response from the remote
		chrome instance describing the problem and it's cause.

		Otherwise, the decoded json data-structure returned from the remote instance is
		returned.

		'''

		self.transport.check_process_ded()
		ret = self.transport.synchronous_command(tab_key=self.tab_id, *args, **kwargs)
		self.transport.check_process_ded()
		self.__check_ret(ret)
		self.transport.check_process_ded()
		return ret

	def drain_transport(self):
		'''
		"Drain" the transport connection.

		This command simply returns all waiting messages sent from the remote chrome
		instance. This can be useful when waiting for a specific asynchronous message
		from chrome, but higher level calls are better suited for managing wait-for-message
		type needs.

		'''
		self.transport.check_process_ded()
		ret = self.transport.drain(tab_key=self.tab_id)
		self.transport.check_process_ded()
		return ret

	def new_tab(self, *args, **kwargs):
		new = self.__class__(use_execution_manager=(self.transport, uuid.uuid4()), *args, **kwargs)
		self.transport.check_process_ded()
		return new

	def close(self):
		self.transport.check_process_ded()
		if self.is_root_session:
			self.transport.close_all()
		else:
			self.transport.close_tab(tab_key=self.tab_id)

		gc.collect()



if __name__ == '__main__':
	import doctest
	doctest.testmod()




""" Client for the Google Chrome browser's remote debugging api.

	> a = Shell(host='localhost', port=92222)


"""
import os
import json
import sys
import socket
import time
import pprint
import logging
import os.path
import requests
import traceback
import signal
import websocket
import subprocess
import distutils.spawn
from . import cr_exceptions

if 'win' in sys.platform:
	import win32con
	import win32api

ACTIVE_PORTS = set()




class ChromeExecutionManager():
	"""
	Class for managing talking to a chromium instance, as well as
	managing the chromium instance's execution lifetime.

	"""

	def __init__(self,
			binary,
			base_tab_key,
			host              = 'localhost',
			port              = None,
			websocket_timeout = 10,
			enable_gpu        = False,
			):
		"""

		Start up a chromium instance binary `binary`, with the base debug port `port`,
		on host `localhost`.

		Note: If the port is set to None (the default), the port will be the next availble port accounting for other
		chrome instances in the current python process.

		If port is not none, and the port is already in use, you will get a ReusedPortError.

		base_tab_key is any hashable python object that is used for multi-tab interfacing.

		"""

		if port is None:
			port = 9222 + (os.getpid() & 0x3FFF)
			if port > 65530:
				port -= 5000
			while port in ACTIVE_PORTS:
				port += 1

		if port in ACTIVE_PORTS:
			raise cr_exceptions.ReusedPortError("Attempting to start chromium using a already-in-use debug port (%s, %s)!" % (port, ACTIVE_PORTS))

		ACTIVE_PORTS.add(port)

		self.binary            = binary
		self.host              = host
		self.port              = port
		self.enable_gpu        = enable_gpu
		self.msg_id            = 0
		self.websocket_timeout = websocket_timeout

		self.tablist = None
		self.soclist = {}
		self.tab_id_map = {}

		self.log = logging.getLogger("Main.ChromeController.ExecutionManager")


		self.log.info("Launching binary %s", self.binary)

		# We retry starting chromium a few times, because it's either brittle
		# or sometimes takes more then 10 seconds to start.
		# Not sure which.
		for x in range(999):
			try:
				self._launch_process(self.binary, self.port, base_tab_key)
				break
			except cr_exceptions.ChromeConnectFailure:
				if x > 3:
					raise

		# self.log.info("Connecting to %s:%s", self.host, self.port)
		# self.connect(base_tab_key)


		self.messages = {}

	def _launch_process(self, binary, dbg_port, base_tab_key):

		if binary is None:
			binary = "chromium"
		if not os.path.exists(binary):
			fixed = distutils.spawn.find_executable(binary)
			if fixed:
				binary = fixed
		if not binary or not os.path.exists(binary):
			raise RuntimeError("Could not find binary '%s'" % binary)

		argv = [
				binary,
				'--headless',
				'--remote-debugging-port={dbg_port}'.format(dbg_port=dbg_port),
				'--enable-features=NetworkService',
			]
		if self.enable_gpu is False:
			argv.append('--disable-gpu')

		# We need a separate process group on windows,
		# to make ctrl+c work properly.
		creationflags = 0
		if 'win' in sys.platform:
			creationflags |= subprocess.CREATE_NEW_PROCESS_GROUP

		preexec_fn = None
		if 'linux' in sys.platform:
			from . import exit_handler
			preexec_fn = exit_handler.on_parent_exit('SIGTERM')

		self.cr_proc = subprocess.Popen(argv,
										stdin         = open(os.path.devnull, "r"),
										stdout        = subprocess.PIPE,
										stderr        = subprocess.PIPE,
										creationflags = creationflags,
										preexec_fn    = preexec_fn,
									)

		self.log.debug("Spawned process: %s, PID: %s", self.cr_proc, self.cr_proc.pid)
		for x in range(100):
			try:
				self.tablist = self.fetch_tablist()

				if not self.tablist:
					raise cr_exceptions.ChromeStartupException("No tabs in started chromium?")

				# self.tab_id_map[base_tab_key] = self.tablist[0]['id']
				# print("Tab base key:", base_tab_key, self.tablist[0]['id'])
				return
			except cr_exceptions.ChromeConnectFailure as e:
				if x > 8:
					raise e
				time.sleep(2)

	def __close_internal_linux(self):
		self.log.debug("Sending sigint to chromium")
		self.cr_proc.send_signal(signal.SIGINT)
		try:
			self.check_process_ded()  # Needed to flush the communcation pipes, sometimes
			self.log.debug("Waiting for chromium to exit")
			try:
				self.cr_proc.wait(timeout=5)
			except subprocess.TimeoutExpired:
				self.log.warning("Chromium failed to exit on sigint")

			try:
				self.cr_proc.terminate()
			except ProcessLookupError:
				self.log.debug("Process exited normally, no need to terminate.")

			self.check_process_ded()  # Processes may dangle until the pipes are closed.

			try:
				self.cr_proc.kill()
			except ProcessLookupError:
				self.log.debug("Process exited normally, no need to terminate.")

			self.check_process_ded()  # Processes may dangle until the pipes are closed.

		except cr_exceptions.ChromeDiedError:
			# Considering we're /trying/ to kill chrome, it
			# dying is OK.
			pass

		self.log.debug("Pid: %s, Return code: %s", self.cr_proc.pid, self.cr_proc.returncode)
		self.log.debug("Chromium closed!")

	def __close_internal_windows(self):
		self.log.debug("Sending CTRL_C_EVENT to chromium")

		win32api.GenerateConsoleCtrlEvent(win32con.CTRL_C_EVENT, self.cr_proc.pid)
		self.cr_proc.send_signal(signal.CTRL_BREAK_EVENT)
		self.cr_proc.send_signal(signal.CTRL_C_EVENT)
		self.cr_proc.send_signal(1)
		self.log.debug("Sent")
		try:
			self.check_process_ded()  # Needed to flush the communcation pipes, sometimes

			# I can't get this to fucking ever work,
			# so just skip and go straight to termination.
			# self.log.debug("Waiting for chromium to exit")
			# try:
			# 	self.cr_proc.wait(timeout=5)
			# except subprocess.TimeoutExpired:
			# 	pass

			try:
				self.cr_proc.terminate()
			except ProcessLookupError:
				self.log.debug("Process exited normally, no need to terminate.")

			self.check_process_ded()  # Processes may dangle until the pipes are closed.

			try:
				self.cr_proc.kill()
			except ProcessLookupError:
				self.log.debug("Process exited normally, no need to terminate.")

			self.check_process_ded()  # Processes may dangle until the pipes are closed.

		except cr_exceptions.ChromeDiedError:
			self.log.debug("ChromeDiedError while polling. Ignoring due to shutdown.")
			# Considering we're /trying/ to kill chrome, it
			# dying is OK.

		self.log.debug("Pid: %s, Return code: %s", self.cr_proc.pid, self.cr_proc.returncode)
		self.log.debug("Chromium closed!")


	def close_chromium(self):
		'''
		Close the remote chromium instance.

		This command is normally executed as part of the class destructor.
		It can be called early without issue, but calling ANY class functions
		after the remote chromium instance is shut down will have unknown effects.

		Note that if you are rapidly creating and destroying ChromeController instances,
		you may need to *explicitly* call this before destruction.
		'''
		if self.cr_proc:
			try:
				if 'win' in sys.platform:
					self.__close_internal_windows()
				else:
					self.__close_internal_linux()
			except Exception as e:
				for line in traceback.format_exc().split("\n"):
					self.log.error(line)



		ACTIVE_PORTS.discard(self.port)


	def check_process_ded(self):
		self.cr_proc.poll()
		if self.cr_proc.returncode != None:
			try:
				stdout, stderr = self.cr_proc.communicate()
				raise cr_exceptions.ChromeDiedError("Chromium process died unexpectedly! Don't know "
					"how to continue!\n	Chromium stdout: {}\n	Chromium stderr: {}".format(stdout.decode("utf-8"), stderr.decode("utf-8")))
			except ValueError:
				# The communcation pipes can go away if the process has exited.
				# If so, ignore the resulting error.
				raise cr_exceptions.ChromeDiedError("Chromium process died unexpectedly! Don't know how to continue!")

	def _get_tab_idx_for_key(self, tab_key):

		if tab_key not in self.tab_id_map:
			return None

		cr_tab_id = self.tab_id_map[tab_key]['id']

		for idx, tab_dat in enumerate(self.tablist):
			if tab_dat['id'] == cr_tab_id:
				return idx

		self.log.error("Missing tab %s from tab list!", tab_key)
		self.log.error("Corresponding cr tab id: %s!", cr_tab_id)
		for tab_key, cr_tab_id in self.tab_id_map.items():
			self.log.error("Tab pair %s -> %s", tab_key, cr_tab_id['id'])
		self.log.error("Chromium tabs (%s):", len(self.tablist))
		for tab in self.tablist:
			self.log.error("	cr tab id: %s", tab['id'])

		raise cr_exceptions.ChromeTabNotFoundError("Tab with ID %s (cr ID: %s) not found!" % (tab_key, cr_tab_id['id']))

	def _get_cr_tab_meta_for_key(self, tab_key):
		if tab_key not in self.tab_id_map:
			return None

		return self.tab_id_map[tab_key]

	def connect(self, tab_key):
		"""
		Open a websocket connection to remote browser, determined by
		self.host and self.port.  Each tab has it's own websocket
		endpoint.

		"""

		assert self.tablist is not None

		tab_idx = self._get_tab_idx_for_key(tab_key)

		if not self.tablist:
			self.tablist = self.fetch_tablist()

		for fails in range(9999):
			try:
				# If we're one past the end of the tablist, we need to create a new tab
				if tab_idx is None:
					self.log.debug("Creating new tab (%s active)", len(self.tablist))
					self.__create_new_tab(tab_key)

				self.__connect_to_tab(tab_key)
				break

			except cr_exceptions.ChromeConnectFailure as e:
				if fails > 6:
					self.log.error("Failed to fetch tab websocket URL after %s retries. Aborting!", fails)
					raise e
				self.log.info("Tab may not have started yet (%s tabs active). Recreating.", len(self.tablist))
				# self.log.info("Tag: %s", self.tablist[tab_idx])


				# For reasons I don't understand, sometimes a new tab doesn't get a websocket
				# debugger URL. Anyways, we close and re-open the tab if that happens.
				# TODO: Handle the case when this happens on the first tab. I think closing the first
				#       tab will kill chromium.
				self.__close_tab(tab_key)

	def __repr__(self):
		return "<ChromeExecutionManager for tabs \n%s\n>" % pprint.pformat(self.tab_id_map)

	def pprint_tablist(self):
		self.log.info("Tablist type: %s", type(self.tablist))
		for line in pprint.pformat(list(enumerate(self.tablist))).split("\n"):
			self.log.info("	%s", line)

	def __create_new_tab(self, tab_key, start_at_url=None):
		url = "http://%s:%s/json/new" % (self.host, self.port)
		if start_at_url:
			url += "?%s" % (start_at_url, )
		try:
			requests.get(url)
			self.tablist = self.fetch_tablist()
			known_tabs = list(self.tab_id_map.values())
			for tab in self.tablist:

				# Glom onto the first tab in the tablist that's not known.
				if tab['id'] not in [tmp['id'] for tmp in known_tabs]:
					self.log.debug("New tab created with ID: '%s'", tab['id'])
					self.tab_id_map[tab_key] = tab
					return


		except requests.exceptions.ConnectionError:
			raise cr_exceptions.ChromeConnectFailure("Failed to create a new tab in remote chromium!")

	def __connect_to_tab(self, tab_key):
		assert tab_key not in self.soclist

		cr_tab_meta = self._get_cr_tab_meta_for_key(tab_key)

		if not 'webSocketDebuggerUrl' in cr_tab_meta:
			raise cr_exceptions.ChromeConnectFailure("Tab %s has no 'webSocketDebuggerUrl' (%s)" % (tab_key, self.tablist))

		wsurl = cr_tab_meta['webSocketDebuggerUrl']

		try:
			self.log.info("Setting up websocket connection for key '%s'", tab_key)
			self.soclist[tab_key] = websocket.create_connection(wsurl)
			self.soclist[tab_key].settimeout(self.websocket_timeout)

		except (socket.timeout, websocket.WebSocketTimeoutException):
			raise cr_exceptions.ChromeCommunicationsError("Could not connect to remote chromium.")

	def __close_tab(self, tab_key, timeout=None):

		# traceback.print_stack()

		cr_tab_id = self.tab_id_map[tab_key]['id']

		url = "http://%s:%s/json/close/%s" % (self.host, self.port, cr_tab_id)
		requests.get(url, timeout=timeout)

		# Delete the now-removed tab from the tab map
		self.tab_id_map.pop(tab_key)

		self.log.info("Closing websocket connecton %s (%s)", tab_key, len(self.soclist))
		self.soclist.pop(tab_key, None)

		self.tablist = self.fetch_tablist()
		return self.tablist

	def close_tab(self, tab_key):
		self.log.info("Closing tab %s (cr ID: %s)", tab_key, self.tab_id_map[tab_key]['id'])
		self.__close_tab(tab_key)

		# If we've closed all the chrome tabs, shut down the interface.
		if not len(self.tab_id_map):
			self.log.info("All tabs are closed. Closing chromium!")
			self.close_websockets()
			self.close_chromium()

	def close_all(self):
		self.log.info("Closing all tabs.")
		for tab_key in list(self.tab_id_map.keys()):
			self.log.info("Closing tab %s (cr ID: %s)", tab_key, self.tab_id_map[tab_key]['id'])
			self.__close_tab(tab_key)

		self.log.info("All tabs are closed. Closing chromium!")
		self.close_websockets()
		self.close_chromium()

	def close_websockets(self):
		""" Close websocket connection to remote browser."""
		self.log.info("Websocket Teardown called")
		for key in list(self.soclist.keys()):
			if self.soclist[key]:
				self.soclist[key].close()
			self.soclist.pop(key)


	def fetch_tablist(self):
		"""Connect to host:port and request list of tabs
			 return list of dicts of data about open tabs."""
		# find websocket endpoint
		try:
			response = requests.get("http://%s:%s/json" % (self.host, self.port))
		except requests.exceptions.ConnectionError:
			raise cr_exceptions.ChromeConnectFailure("Failed to fetch configuration json from browser!")

		tablist = json.loads(response.text)

		return tablist

	def __check_open_socket(self, tab_key):
		# self.log.info("__check_open_socket -> %s", tab_key)
		if not tab_key in self.soclist:
			self.connect(tab_key=tab_key)
		if self.soclist[tab_key].connected is not True:
			self.connect(tab_key=tab_key)
		if not tab_key in self.messages:
			self.messages[tab_key] = []


	def synchronous_command(self, command, tab_key, **params):
		"""
		Synchronously execute command `command` with params `params` in the
		remote chrome instance, returning the response from the chrome instance.

		"""
		self.log.debug("Synchronous_command to tab %s (%s):", tab_key, self._get_cr_tab_meta_for_key(tab_key))
		self.log.debug("	command: '%s'", command)
		self.log.debug("	params:  '%s'", params)
		self.log.debug("	tab_key:  '%s'", tab_key)

		send_id = self.send(command=command, tab_key=tab_key, params=params)
		resp = self.recv(message_id=send_id, tab_key=tab_key)

		self.log.debug("	Response: '%s'", str(resp).encode("ascii", 'ignore').decode("ascii"))

		# self.log.debug("	resolved tab idx %s:", self.tab_id_map[tab_key])
		return resp

	def send(self, command, tab_key, params=None):
		'''
		Send command `command` with optional parameters `params` to the
		remote chrome instance.

		The command `id` is automatically added to the outgoing message.

		return value is the command id, which can be used to match a command
		to it's associated response.
		'''
		self.__check_open_socket(tab_key)

		sent_id = self.msg_id

		command = {
				"id": self.msg_id,
				"method": command,
			}

		if params:
			command["params"] = params
		navcom = json.dumps(command)


		# self.log.debug("		Sending: '%s'", navcom)
		try:
			self.soclist[tab_key].send(navcom)
		except (socket.timeout, websocket.WebSocketTimeoutException):
			raise cr_exceptions.ChromeCommunicationsError("Failure sending command to chromium.")
		except websocket.WebSocketConnectionClosedException:
			raise cr_exceptions.ChromeCommunicationsError("Websocket appears to have been closed. Is the"
				" remote chromium instance dead?")


		self.msg_id += 1
		return sent_id


	def ___recv(self, tab_key, timeout=None):

		try:
			if timeout:
				self.soclist[tab_key].settimeout(timeout)

			tmp = self.soclist[tab_key].recv()
			self.log.debug("		Received: '%s'", tmp)

			decoded = json.loads(tmp)
			return decoded
		except (socket.timeout, websocket.WebSocketTimeoutException):
			return None
		except websocket.WebSocketConnectionClosedException:
			raise cr_exceptions.ChromeCommunicationsError("Websocket appears to have been closed. Is the"
				" remote chromium instance dead?")
		finally:
			self.soclist[tab_key].settimeout(self.websocket_timeout)

	def recv_filtered(self, keycheck, tab_key, timeout=30, message=None):
		'''
		Receive a filtered message, using the callable `keycheck` to filter received messages
		for content.

		`keycheck` is expected to be a callable that takes a single parameter (the decoded response
		from chromium), and returns a boolean (true, if the command is the one filtered for, or false
		if the command is not the one filtered for).

		This is used internally, for example, by `recv()`, to filter the response for a specific ID:

		```
			def check_func(message):
				if message_id is None:
					return True
				if "id" in message:
					return message['id'] == message_id
				return False
			return self.recv_filtered(check_func, timeout)

		```

		Note that the function is defined dynamically, and `message_id` is captured via closure.

		'''


		self.__check_open_socket(tab_key)

		# First, check if the message has already been received.
		for idx in range(len(self.messages[tab_key])):
			if keycheck(self.messages[tab_key][idx]):
				return self.messages[tab_key].pop(idx)

		timeout_at = time.time() + timeout
		while 1:
			tmp = self.___recv(tab_key)
			if keycheck(tmp):
				return tmp
			else:
				self.messages[tab_key].append(tmp)

			if time.time() > timeout_at:
				if message:
					raise cr_exceptions.ChromeResponseNotReceived("Failed to receive response in recv_filtered() (%s)" % message)
				else:
					raise cr_exceptions.ChromeResponseNotReceived("Failed to receive response in recv_filtered()")
			else:
				time.sleep(0.005)

	def recv_all_filtered(self, keycheck, tab_key, timeout=0.5):
		'''
		Receive a all messages matching a filter, using the callable `keycheck` to filter received messages
		for content.

		This function will *ALWAY* block for at least `timeout` seconds.

		If chromium is for some reason continuously streaming responses, it may block forever!

		`keycheck` is expected to be a callable that takes a single parameter (the decoded response
		from chromium), and returns a boolean (true, if the command is the one filtered for, or false
		if the command is not the one filtered for).

		```
			def check_func(message):
				if message_id is None:
					return True
				if "id" in message:
					return message['id'] == message_id
				return False
			return self.recv_filtered(check_func, timeout)

		```

		Note that the function is defined dynamically, and `message_id` is captured via closure.

		'''


		self.__check_open_socket(tab_key)
		# First, check if the message has already been received.
		ret           = [tmp for tmp in self.messages[tab_key] if keycheck(tmp)]
		self.messages[tab_key] = [tmp for tmp in self.messages[tab_key] if not keycheck(tmp)]

		self.log.debug("Waiting for all messages from the socket")
		timeout_at = time.time() + timeout
		while 1:
			tmp = self.___recv(tab_key, timeout=timeout)
			if keycheck(tmp):
				ret.append(tmp)
			else:
				self.messages[tab_key].append(tmp)

			if time.time() > timeout_at:
				return ret
			else:
				self.log.debug("Sleeping: %s, %s" % (timeout_at, time.time()))
				time.sleep(0.005)

	def recv(self, tab_key, message_id=None, timeout=30):
		'''
		Recieve a message, optionally filtering for a specified message id.

		If `message_id` is none, the first command in the receive queue is returned.
		If `message_id` is not none, the command waits untill a message is received with
		the specified id, or it times out.

		Timeout is the number of seconds to wait for a response, or `None` if the timeout
		has expired with no response.

		'''

		self.__check_open_socket(tab_key)

		# First, check if the message has already been received.
		for idx in range(len(self.messages[tab_key])):
			if self.messages[tab_key][idx]:
				if "id" in self.messages[tab_key][idx] and message_id:
					if self.messages[tab_key][idx]['id'] == message_id:
						return self.messages[tab_key].pop(idx)

		# Then spin untill we either have the message,
		# or have timed out.
		def check_func(message):
			if message_id is None:
				return True
			if not message:
				self.log.debug("Message is not true (%s)!", message)
				return False
			if "id" in message:
				return message['id'] == message_id
			return False

		return self.recv_filtered(check_func, tab_key, timeout)

	def flush(self, tab_key):
		'''
		Flush the pending RX buffer for a specific tab key.
		'''
		self.messages[tab_key] = []



	def drain(self, tab_key):
		'''
		Return all messages in waiting for the websocket connection.
		'''
		self.log.debug("Draining transport")
		ret = []
		while len(self.messages[tab_key]):
			ret.append(self.messages[tab_key].pop(0))

		self.log.debug("Polling socket")

		tmp = self.___recv(tab_key)
		while tmp is not None:
			ret.append(tmp)
			tmp = self.___recv(tab_key)

		self.log.debug("Drained %s messages", len(ret))
		return ret

	def __del__(self):
		self.log.debug("Transport destructor called. Tearing down chromium")
		try:
			self.close_chromium()
		except Exception:
			pass

		ACTIVE_PORTS.discard(self.port)

if __name__ == '__main__':
	import doctest
	doctest.testmod()




import unittest
import socket
import json
import base64
import zlib
import gzip
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread

import ChromeController
from . import testing_server

CHROME_BINARY_NAME = "google-chrome"
TIMEOUT_SECS       = 5

class TestPlainCreation(unittest.TestCase):
	def test_plain_instantiation_1(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			self.assertTrue(cr is not None)




class TestSimpleFetch(unittest.TestCase):
	def setUp(self):

		# Configure mock server.
		self.mock_server_port, self.mock_server, self.mock_server_thread = testing_server.start_server(self, {})

	def tearDown(self):
		self.mock_server.shutdown()
		self.mock_server_thread.join()

	def test_fetch_1(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			resp = cr.blocking_navigate_and_get_source("http://localhost:{}".format(self.mock_server_port), timeout=TIMEOUT_SECS)
			self.assertEqual(resp['content'], 'Root OK?')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/html")

	def test_fetch_decode_1(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			# text/html content should be decoded automatically.
			resp = cr.blocking_navigate_and_get_source("http://localhost:{}/html-decode".format(self.mock_server_port), timeout=TIMEOUT_SECS)
			self.assertEqual(resp['content'], 'Root OK?')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/html")

	def test_fetch_decode_json_1(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			resp = cr.blocking_navigate_and_get_source("http://localhost:{}/json/valid".format(self.mock_server_port), timeout=TIMEOUT_SECS)
			self.assertEqual(resp['content'], '{"oh" : "hai"}')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "application/json")

	def test_fetch_decode_json_2(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			resp = cr.blocking_navigate_and_get_source("http://localhost:{}/json/no-coding".format(self.mock_server_port), timeout=TIMEOUT_SECS)
			self.assertEqual(resp['content'], '{"oh" : "hai"}')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/plain")

	def test_fetch_decode_json_3(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			resp = cr.blocking_navigate_and_get_source("http://localhost:{}/json/invalid".format(self.mock_server_port), timeout=TIMEOUT_SECS)
			self.assertEqual(resp['content'], 'LOLWAT')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "application/json")

	def test_fetch_compressed_1(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			resp = cr.blocking_navigate_and_get_source("http://localhost:{}/compressed/gzip".format(self.mock_server_port), timeout=TIMEOUT_SECS)
			self.assertEqual(resp['content'], 'Root OK?')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/html")

	def test_fetch_compressed_2(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			resp = cr.blocking_navigate_and_get_source("http://localhost:{}/compressed/deflate".format(self.mock_server_port), timeout=TIMEOUT_SECS)
			self.assertEqual(resp['content'], 'Root OK?')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/html")

	########################################################################################################################
	########################################################################################################################
	########################################################################################################################

	def test_get_head_1(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			inurl_1 = "http://localhost:{}/".format(self.mock_server_port)

			cr.blocking_navigate(inurl_1)

			_, nurl_1 = cr.get_page_url_title()
			nurl2_1 = cr.get_current_url()

			self.assertEqual(nurl2_1, inurl_1)
			self.assertEqual(nurl_1,  inurl_1)

	# Binary shit is broken.
	# Siiigh.
	# def test_get_head_2(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		inurl_2 = "http://localhost:{}/filename_mime/content-disposition".format(self.mock_server_port)

	# 		cr.blocking_navigate(inurl_2)

	# 		nurl_2, _ = cr.get_page_url_title()
	# 		nurl2_2 = cr.get_current_url()

	# 		self.assertEqual(inurl_2, nurl_2)
	# 		self.assertEqual(inurl_2, nurl2_2)

	########################################################################################################################

	def test_redirect_handling_1(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:

			inurl_1 = "http://localhost:{}/redirect/from-1".format(self.mock_server_port)
			resp = cr.blocking_navigate_and_get_source(inurl_1, timeout=TIMEOUT_SECS)

			self.assertEqual(resp['content'], 'Redirect-To-1')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/html")

	# def test_redirect_handling_2(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		inurl_2 = "http://localhost:{}/redirect/from-2".format(self.mock_server_port)
	# 		resp = cr.blocking_navigate_and_get_source(inurl_2, timeout=TIMEOUT_SECS)

	# 		self.assertEqual(resp['content'], 'Redirect-To-2')
	# 		self.assertEqual(resp['binary'], False)
	# 		self.assertEqual(resp['mimetype'], "text/html")

	def test_redirect_handling_3(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			inurl_3 = "http://localhost:{}/redirect/from-3".format(self.mock_server_port)
			resp = cr.blocking_navigate_and_get_source(inurl_3, timeout=TIMEOUT_SECS)

			self.assertEqual(resp['content'], 'Redirect-To-3')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/html")

	def test_redirect_handling_4(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			inurl_3 = "http://localhost:{}/redirect/from-4".format(self.mock_server_port)
			resp = cr.blocking_navigate_and_get_source(inurl_3, timeout=TIMEOUT_SECS)

			self.assertEqual(resp['content'], 'Redirect-To-4')
			self.assertEqual(resp['binary'], False)
			self.assertEqual(resp['mimetype'], "text/html")

	def test_redirect_handling_3(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			inurl_3 = "http://localhost:{}/redirect/from-1".format(self.mock_server_port)
			outurl_3 = "http://localhost:{}/redirect/to-1".format(self.mock_server_port)

			cr.blocking_navigate(inurl_3)
			_, nurl_3 = cr.get_page_url_title()
			nurl2_3 = cr.get_current_url()

			self.assertEqual(outurl_3, nurl_3)
			self.assertEqual(outurl_3, nurl2_3)

	# Broken
	# def test_redirect_handling_4(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		inurl_4 = "http://localhost:{}/redirect/from-2".format(self.mock_server_port)
	# 		outurl_4 = "http://localhost:{}/redirect/to-2".format(self.mock_server_port)

	# 		cr.blocking_navigate(inurl_4)
	# 		_, nurl_4 = cr.get_page_url_title()
	# 		nurl2_4 = cr.get_current_url()
	# 		self.assertEqual(outurl_4, nurl_4)
	# 		self.assertEqual(outurl_4, nurl2_4)

	# Broken too.
	# def test_redirect_handling_5(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		# This is a redirect without the actual redirect
	# 		with self.assertRaises(ValueError):
	# 			inurl_5 = "http://localhost:{}/redirect/bad-1".format(self.mock_server_port)
	# 			cr.blocking_navigate(inurl_5)

	def test_redirect_handling_6(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			# This is a infinitely recursive redirect.
			with self.assertRaises(ChromeController.ChromeResponseNotReceived):
				inurl_6 = "http://localhost:{}/redirect/bad-2".format(self.mock_server_port)
				cr.blocking_navigate(inurl_6)

	def test_redirect_handling_7(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			# This is a infinitely recursive redirect.
			with self.assertRaises(ChromeController.ChromeResponseNotReceived):
				inurl_6 = "http://localhost:{}/redirect/bad-3".format(self.mock_server_port)
				cr.blocking_navigate(inurl_6)

	def test_redirect_handling_8(self):
		with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
			inurl_7 = "http://localhost:{}/redirect/from-5".format(self.mock_server_port)
			# Assumes localhost seems to resolve to the listening address (here it's 0.0.0.0). Is this ever not true? IPv6?
			outurl_7 = "http://127.0.0.1:{}/".format(self.mock_server_port)

			cr.blocking_navigate(inurl_7)
			_, nurl_7 = cr.get_page_url_title()
			nurl2_7 = cr.get_current_url()
			self.assertEqual(outurl_7, nurl_7)
			self.assertEqual(outurl_7, nurl2_7)


	# def test_get_item_1(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		inurl_1 = "http://localhost:{}".format(self.mock_server_port)
	# 		content_1, fileN_1, mType_1 = self.wg.getItem(inurl_1)
	# 		self.assertEqual(content_1, 'Root OK?')
	# 		self.assertEqual(fileN_1, '')
	# 		self.assertEqual(mType_1, "text/html")


	# def test_get_item_2(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		inurl_2 = "http://localhost:{}/filename_mime/content-disposition".format(self.mock_server_port)
	# 		content_2, fileN_2, mType_2 = self.wg.getItem(inurl_2)

	# 		# Lack of an explicit mimetype makes this not get decoded
	# 		self.assertEqual(content_2, b'LOLWAT?')
	# 		self.assertEqual(fileN_2, 'lolercoaster.txt')
	# 		self.assertEqual(mType_2, None)


	# def test_get_item_3(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		inurl_3 = "http://localhost:{}/filename/path-only.txt".format(self.mock_server_port)
	# 		content_3, fileN_3, mType_3 = self.wg.getItem(inurl_3)

	# 		self.assertEqual(content_3, b'LOLWAT?')
	# 		self.assertEqual(fileN_3, 'path-only.txt')
	# 		self.assertEqual(mType_3, None)

	# def test_get_cookies_1(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		inurl_1 = "http://localhost:{}/cookie_test".format(self.mock_server_port)
	# 		inurl_2 = "http://localhost:{}/cookie_require".format(self.mock_server_port)

	# 		self.wg.clearCookies()
	# 		cookies = self.wg.getCookies()
	# 		self.assertEqual(list(cookies), [])

	# 		page_resp_nocook = cr.blocking_navigate_and_get_source(inurl_2, timeout=TIMEOUT_SECS)
	# 		self.assertEqual(page_resp_nocook, '<html><body>Cookie is missing</body></html>')


	# 		_ = cr.blocking_navigate_and_get_source(inurl_1, timeout=TIMEOUT_SECS)
	# 		cookies = self.wg.getCookies()
	# 		print(cookies)

	# 		page_resp_cook = cr.blocking_navigate_and_get_source(inurl_2, timeout=TIMEOUT_SECS)
	# 		self.assertEqual(page_resp_cook, '<html><body>Cookie forwarded properly!</body></html>')




	########################################################################################################################
	########################################################################################################################
	########################################################################################################################

	# # So binary downloads are broken ATM
	# # see https://bugs.chromium.org/p/chromium/issues/detail?id=831887.
	# # Siiiiigh.

	# def test_file_and_name_1(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source("http://localhost:{}/filename/path-only.txt".format(self.mock_server_port), timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		# self.assertEqual(fn, 'path-only.txt')


	# def test_file_and_name_2(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source("http://localhost:{}/filename/content-disposition".format(self.mock_server_port), timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, 'lolercoaster.txt')


	# def test_file_and_name_3(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source("http://localhost:{}/filename_mime/content-disposition-quotes-1".format(self.mock_server_port), timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, 'lolercoaster.html')


	# def test_file_and_name_4(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source("http://localhost:{}/filename_mime/content-disposition-quotes-2".format(self.mock_server_port), timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, 'lolercoaster.html')


	# def test_file_and_name_5(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source("http://localhost:{}/filename_mime/content-disposition-quotes-spaces-1".format(self.mock_server_port), timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, 'loler coaster.html')


	# def test_file_and_name_6(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source("http://localhost:{}/filename_mime/content-disposition-quotes-spaces-2".format(self.mock_server_port), timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, 'loler coaster.html')


	# def test_file_and_name_7(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source(requestedUrl="http://localhost:{}/filename_mime/content-disposition-quotes-spaces-2".format(self.mock_server_port), timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, 'loler coaster.html')

	# def test_file_and_name_8(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source(requestedUrl="http://localhost:{}/filename_mime/content-disposition-quotes-spaces-2".format(self.mock_server_port), addlHeaders={"Referer" : 'http://www.example.org'}, timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, 'loler coaster.html')

	# def test_file_and_name_9(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source("http://localhost:{}/filename_mime/content-disposition-quotes-spaces-2".format(self.mock_server_port), addlHeaders={"Referer" : 'http://www.example.org'}, timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, 'loler coaster.html')

	# def test_file_and_name_10(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source("http://localhost:{}/filename/path-only-trailing-slash/".format(self.mock_server_port), timeout=TIMEOUT_SECS)
	# 		self.assertEqual(resp['content'], b"LOLWAT?\x00\xff\xaa\x55!")
	# 		self.assertEqual(resp['binary'], True)
	# 		# self.assertEqual(fn, '')


	# def test_file_name_mime_1(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source(
	# 						"http://localhost:{}/filename_mime/path-only.txt".format(self.mock_server_port))
	# 		self.assertEqual(resp['content'], 'LOLWAT?')
	# 		self.assertEqual(resp['binary'], False)
	# 		# self.assertEqual(fn, 'path-only.txt')
	# 		self.assertEqual(resp['mimetype'], 'text/plain')

	# def test_file_name_mime_2(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source(
	# 						"http://localhost:{}/filename_mime/content-disposition".format(self.mock_server_port))
	# 		self.assertEqual(resp['content'], 'LOLWAT?')
	# 		self.assertEqual(resp['binary'], False)
	# 		# self.assertEqual(fn, 'lolercoaster.txt')
	# 		self.assertEqual(resp['mimetype'], 'text/plain')

	# def test_file_name_mime_3(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source(
	# 						"http://localhost:{}/filename_mime/content-disposition-html-suffix".format(self.mock_server_port))
	# 		self.assertEqual(resp['content'], 'LOLWAT?')
	# 		self.assertEqual(resp['binary'], False)
	# 		# self.assertEqual(fn, 'lolercoaster.html')
	# 		self.assertEqual(resp['mimetype'], 'text/plain')

	# def test_file_name_mime_5(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source(
	# 						"http://localhost:{}/filename/path-only-trailing-slash/".format(self.mock_server_port))
	# 		self.assertEqual(resp['content'], 'LOLWAT?')
	# 		self.assertEqual(resp['binary'], False)
	# 		# self.assertEqual(fn, '')
	# 		self.assertEqual(resp['mimetype'], 'text/plain')

	# def test_file_name_mime_4(self):
	# 	with ChromeController.ChromeContext(binary=CHROME_BINARY_NAME) as cr:
	# 		resp = cr.blocking_navigate_and_get_source(
	# 						"http://localhost:{}/filename_mime/explicit-html-mime".format(self.mock_server_port))
	# 		self.assertEqual(resp['content'], 'LOLWAT?')
	# 		self.assertEqual(resp['binary'], False)
	# 		# self.assertEqual(fn, 'lolercoaster.html')
	# 		self.assertEqual(resp['mimetype'], 'text/html')



"""
"""
import json
import logging
import ast
import difflib
import sys
import pprint
import os.path

import astor


CHECKS = {
	"array"   : ast.Tuple(ctx=ast.Load(), elts=[ast.Name(id='list', ctx=ast.Load()), ast.Name(id='tuple', ctx=ast.Load())]),   # "(list, tuple)",
	"boolean" : ast.Tuple(ctx=ast.Load(), elts=[ast.Name(id='bool', ctx=ast.Load())]),   # "(bool, )",
	"integer" : ast.Tuple(ctx=ast.Load(), elts=[ast.Name(id='int', ctx=ast.Load())]),   # "(int, )",
	"number"  : ast.Tuple(ctx=ast.Load(), elts=[ast.Name(id='float', ctx=ast.Load()), ast.Name(id='int', ctx=ast.Load())]),   # "(float, int)",
	"string"  : ast.Tuple(ctx=ast.Load(), elts=[ast.Name(id='str', ctx=ast.Load())]),   # "(str, )",
}



class JsonInterfaceGenerator(object):
	"""

	"""

	def __init__(self, protocol_major=1, protocol_minor=2, debug_prints=False, *args, **kwargs):
		""" init """

		super().__init__(*args, **kwargs)

		self.line_num = 0
		protocol_major = str(protocol_major)
		protocol_minor = str(protocol_minor)

		self.do_debug_prints = debug_prints

		self.types = {}
		self.protocol = self.__load_protocol(protocol_major, protocol_minor)

		self.__build_interface_class()

	def __load_json_file(self, fname):

		folder = os.path.split(__file__)[0]
		protocol_file_path = os.path.join(folder, "../", 'protocols', fname)
		protocol_file_path = os.path.abspath(protocol_file_path)
		assert(os.path.exists(protocol_file_path)), "Protocol file '{}' appears to be missing!".format(protocol_file_path)

		with open(protocol_file_path) as fp:
			protocol_str = fp.read()
		return json.loads(protocol_str)


	def __load_protocol(self, major, minor):

		protocol_rev = "{}.{}".format(major, minor)
		main_json_file = "browser_protocol-r{}.json".format(protocol_rev)
		js_json_file = "js_protocol-r{}.json".format(protocol_rev)

		js_file_1 = self.__load_json_file(main_json_file)
		js_file_2 = self.__load_json_file(js_json_file)

		self.__validate_protocol_version(main_json_file, js_file_1, major, minor)
		self.__validate_protocol_version(js_json_file, js_file_2, major, minor)

		# assemble the two json files into the single command descriptor file.
		for domain in js_file_2['domains']:
			js_file_1['domains'].append(domain)

		return js_file_1

	def __get_line(self):
		self.line_num += 1
		return self.line_num


	def __validate_protocol_version(self, filename, js_file, major, minor):
		errm_1 = "Major version mismatch: {} - {} in file {}".format(js_file['version']["major"], major, filename)
		errm_2 = "Minor version mismatch: {} - {} in file {}".format(js_file['version']["minor"], minor, filename)

		v_1 = js_file['version']["major"]
		v_2 = js_file['version']["minor"]

		assert isinstance(major, str)
		assert isinstance(minor, str)

		assert v_1 == major, errm_1
		assert v_2 == minor, errm_2


	def __build_interface_class(self):
		# body = ast.
		body = [
			ast.Expr(value=ast.Str(s='\n\n\t')),
			self.__build__init()
		]
		for subdom in self.protocol['domains']:
			subdom_funcs = self.__build_domain_interface(subdom)
			body += subdom_funcs

		# print(body)

		self.interface_class = ast.ClassDef(
				name           = "ChromeRemoteDebugInterface",
				bases          = [ast.Name(id="ChromeInterface", ctx=ast.Load())],
				body           = body,
				keywords       = [],
				decorator_list = [],
				starargs       = None,
				kwargs         = None,
				lineno         = self.__get_line(),
				col_offset     = 0,
				)

		# code = astor.dump_tree(self.interface_class)
		# print(code)

	def __build__init(self):

		super_func_call = ast.Call(func=ast.Name(id='super', ctx=ast.Load()), args=[], keywords=[])
		if (sys.version_info[0], sys.version_info[1]) == (3, 5) or \
			(sys.version_info[0], sys.version_info[1]) == (3, 6) or \
			(sys.version_info[0], sys.version_info[1]) == (3, 7):
			super_func = ast.Call(
									func=ast.Attribute(value=super_func_call, attr='__init__', ctx=ast.Load()),
									args=[ast.Starred(value=ast.Name(id='args', ctx=ast.Load()), ctx=ast.Load())],
									keywords=[ast.keyword(arg=None, value=ast.Name(id='kwargs', ctx=ast.Load()), ctx=ast.Load())],
							)
		elif (sys.version_info[0], sys.version_info[1]) == (3,4):
			super_func = ast.Call(
									func=ast.Attribute(value=super_func_call, attr='__init__', ctx=ast.Load()),
									args=[],
									keywords=[],
									starargs=ast.Name(id='args', ctx=ast.Load()),
									kwargs=ast.Name(id='kwargs', ctx=ast.Load()),

							)
		else:
			print("Version:", sys.version_info)
			raise RuntimeError("This script only functions on python 3.4, 3.5, 3.6, or 3.7. Active python version {}.{}".format(*sys.version_info))

		super_init = ast.Expr(
							value=super_func,
							lineno     = self.__get_line(),
							col_offset = 0,
						)

		body = [super_init]

		sig = ast.arguments(
					args=[ast.arg('self', None)],
					vararg=ast.arg(arg='args', annotation=None),
					kwarg=ast.arg(arg='kwargs', annotation=None),
					varargannotation=None,
					kwonlyargs=[],
					kwargannotation=None,
					defaults=[],
					kw_defaults=[])

		func = ast.FunctionDef(
			name = "__init__",
			args = sig,
			body = body,
			decorator_list = [],
			lineno     = self.__get_line(),
			col_offset = 0,
			)

		return func

	def __build_domain_interface(self, subdom):
		assert "domain" in subdom

		dom_desc  = subdom.get("descripton", "")
		dom_name  = subdom['domain']
		full_name = subdom['domain']

		for typen in subdom.get('types', []):
			typestr = "{}_{}".format(dom_name, typen['id'])
			assert typen['id'] not in self.types, "Duplicate type name: {}".format(typen['id'])
			self.types[typestr] = typen

		functions = []
		for command in subdom.get('commands', []):
			func = self.__build_function(dom_name, full_name, command)
			functions.append(func)

		return functions

	def __build_desc_string(self, dom_name, func_name, func_params):
		desc = []
		fname = "{}.{}".format(dom_name, func_name)
		desc.append("Function path: {}".format(fname))
		desc.append("	Domain: {}".format(dom_name))
		desc.append("	Method name: {}".format(func_name))
		desc.append("")
		if 'experimental' in func_params and func_params['experimental']:
			desc.append("	WARNING: This function is marked 'Experimental'!")
			desc.append("")


		if "parameters" in func_params:
			desc.append("	Parameters:")
			required = [param for param in func_params['parameters'] if not param.get("optional", False)]
			optional = [param for param in func_params['parameters'] if param.get("optional", False)]
			sections = [
				("		Required arguments:", required),
				("		Optional arguments:", optional),
			]
			sections = [section for section in sections if section[1]]
			for segment_name, items in sections:
				desc.append(segment_name)
				for param in items:
					if not "description" in param:
						param['description'] = "No description"
					if "type" in param:
						desc.append("			\'{}\' (type: {}) -> {}".format(param['name'], param['type'], param['description']))
					else:
						desc.append("			\'{}\' (type: {}) -> {}".format(param['name'], param['$ref'], param['description']))

		if "returns" in func_params:
			desc.append("	Returns:")
			for param in func_params['returns']:
				if not "description" in param:
					param['description'] = "No description"
				if "type" in param:
					desc.append("		\'{}\' (type: {}) -> {}".format(param['name'], param['type'], param['description']))
				else:
					desc.append("		\'{}\' (type: {}) -> {}".format(param['name'], param['$ref'], param['description']))
		else:
			desc.append("	No return value.")

		desc.append("")

		if "description" in func_params:
			desc.append("	Description: {}".format(func_params['description']))

		desc = ["\t\t"+line for line in desc]
		ret = "\n".join(desc)

		return ret

	def __build_conditional_arg_check(self, argname, argtype):

		target_value = ast.Subscript(
							value=ast.Name(id='kwargs', ctx=ast.Load()),
							slice=ast.Index(ast.Str(s=argname)),
							ctx=ast.Load()
							)

		presence_check = ast.Call(func = ast.Name(id='isinstance', ctx=ast.Load()),
				args         = [target_value, argtype],
				keywords     = [],
				lineno       = self.__get_line())

		# Assumes that argtype is a ast.Tuple of ast.Name items
		types = [t.id for t in argtype.elts]

		check_message = ast.BinOp(
				left         = ast.Str(s='Optional argument \'{}\' must be of type \'{}\'. Received type: \'%s\''.format(argname, types)),
				op           = ast.Mod(),
				right        = ast.Call(func=ast.Name(id='type', ctx=ast.Load()), args=[target_value], keywords=[]),
				lineno       = self.__get_line())

		assert_check = ast.Assert(
			test         = presence_check,
			msg          = check_message,
			lineno       = self.__get_line())

		check_body = [assert_check]

		check = ast.Compare(left=ast.Str(s=argname, ctx=ast.Load()), ops=[ast.In()], comparators=[ast.Name(id='kwargs', ctx=ast.Load())])

		new_ret = ast.If(
			test   = check,
			body   = check_body,
			orelse = [],
			lineno = self.__get_line())

		return new_ret

	def __build_unconditional_arg_check(self, argname, argtype):

		presence_check = ast.Call(func = ast.Name(id='isinstance', ctx=ast.Load()),
				args         = [ast.Name(id=argname, ctx=ast.Load()), argtype],
				keywords     = [],
				lineno       = self.__get_line())

		types = [t.id for t in argtype.elts]
		check_message = ast.BinOp(
				left         = ast.Str(s='Argument \'{}\' must be of type \'{}\'. Received type: \'%s\''.format(argname, types)),
				op           = ast.Mod(),
				right        = ast.Call(func=ast.Name(id='type', ctx=ast.Load()), args=[ast.Name(id=argname, ctx=ast.Load())], keywords=[]),
				lineno       = self.__get_line())

		new_ret = ast.Assert(
			test         = presence_check,
			msg          = check_message,
			lineno       = self.__get_line())


		return new_ret

	def __build_debug_print(self, prefix_str, var_name):
		pstmt = ast.Expr(
			value=ast.Call(
				func     = ast.Name(id='print', ctx=ast.Load()),
				args     = [ast.Str(s=prefix_str), ast.Name(id=var_name, ctx=ast.Load())],
				keywords = [],
				lineno   = self.__get_line()
			)
		)
		return pstmt

	def __build_function(self, dom_name, full_name, func_params):

		assert 'name' in func_params
		func_name = func_params['name']

		docstr = self.__build_desc_string(dom_name, func_name, func_params)

		args = [ast.arg('self', None)]
		message_params = []
		func_body = []

		if docstr:
			func_body.append(ast.Expr(ast.Str("\n"+docstr+"\n\t\t")))

		for param in func_params.get("parameters", []):

			argname = param['name']


			param_optional = param.get("optional", False)

			if param_optional is False:
				message_params.append(ast.keyword(argname, ast.Name(id=argname, ctx=ast.Load())))
				args.append(ast.arg(argname, None))
				if self.do_debug_prints:
					func_body.append(self.__build_debug_print(argname, argname))



			param_type = param.get("type", None)
			if param_type in CHECKS:
				if param_optional:
					check = self.__build_conditional_arg_check(argname, CHECKS[param_type])
				else:
					check = self.__build_unconditional_arg_check(argname, CHECKS[param_type])

				if check:
					func_body.append(check)




		optional_params = [param.get("name") for param in func_params.get("parameters", []) if param.get("optional", False)]
		func_kwargs = None
		if len(optional_params):


			value = ast.List(elts=[ast.Str(s=param, ctx=ast.Store()) for param in optional_params], ctx=ast.Load())
			create_list = ast.Assign(targets=[ast.Name(id='expected', ctx=ast.Store())], value=value)

			func_body.append(create_list)

			passed_arg_list = ast.Assign(targets=[ast.Name(id='passed_keys', ctx=ast.Store())],
				value=ast.Call(func=ast.Name(id='list', ctx=ast.Load()),
				args=[ast.Call(func=ast.Attribute(value=ast.Name(id='kwargs', ctx=ast.Load()), attr='keys', ctx=ast.Load()), args=[], keywords=[])],
				keywords=[]))

			func_body.append(passed_arg_list)

			comprehension = ast.comprehension(target=ast.Name(id='key', ctx=ast.Store()), iter=ast.Name(id='passed_keys', ctx=ast.Load()), ifs=[], is_async=False)
			comparator = ast.Name(id='expected', ctx=ast.Load())

			listcomp = ast.ListComp(elt=ast.Compare(left=ast.Name(id='key', ctx=ast.Load()), ops=[ast.In()], comparators=[comparator]), generators=[comprehension])

			check_message = ast.BinOp(
					left         = ast.Str(s="Allowed kwargs are {}. Passed kwargs: %s".format(optional_params)),
					op           = ast.Mod(),
					right        = ast.Name(id='passed_keys', ctx=ast.Load()),
					lineno       = self.__get_line())

			kwarg_check = ast.Assert(test=ast.Call(func=ast.Name(id='all', ctx=ast.Load()), args=[listcomp], keywords=[]), msg=check_message)
			func_body.append(kwarg_check)

			func_kwargs = ast.Name(id='kwargs', ctx=ast.Load())


		fname = "{}.{}".format(dom_name, func_name)
		fname = ast.Str(s=fname, ctx=ast.Load())


		if (sys.version_info[0], sys.version_info[1]) == (3, 5) or \
			(sys.version_info[0], sys.version_info[1]) == (3, 6) or \
			(sys.version_info[0], sys.version_info[1]) == (3, 7):

			# More irritating minor semantic differences in the AST between 3.4 and 3.5
			if func_kwargs:
				message_params.append(ast.keyword(arg=None, value=ast.Name(id='kwargs', ctx=ast.Load())))

			communicate_call = ast.Call(
					func=ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), ctx=ast.Load(), attr='synchronous_command'),
					args=[fname],
					keywords=message_params)

		elif (sys.version_info[0], sys.version_info[1]) == (3,4):

			communicate_call = ast.Call(
					func=ast.Attribute(value=ast.Name(id='self', ctx=ast.Load()), ctx=ast.Load(), attr='synchronous_command'),
					args=[fname],
					kwargs=func_kwargs,
					keywords=message_params)
		else:
			print("Version:", sys.version_info)
			raise RuntimeError("This script only functions on python 3.4, 3.5, 3.6, or 3.7. Active python version {}.{}".format(*sys.version_info))


		do_communicate = ast.Assign(targets=[ast.Name(id='subdom_funcs', ctx=ast.Store())], value=communicate_call)
		func_ret = ast.Return(value=ast.Name(id='subdom_funcs', ctx=ast.Load()))


		if len(optional_params) and self.do_debug_prints:
			func_body.append(self.__build_debug_print('kwargs', 'kwargs'))

		func_body.append(do_communicate)
		func_body.append(func_ret)

		if len(optional_params):
			kwarg = ast.arg(arg='kwargs', annotation=None)
		else:
			kwarg = None


		sig = ast.arguments(
					args=args,
					vararg=None,
					varargannotation=None,
					kwonlyargs=[],
					kwarg=kwarg,
					kwargannotation=None,
					defaults=[],
					kw_defaults=[])

		func = ast.FunctionDef(
			name = "{}_{}".format(full_name, func_name),
			args = sig,
			body = func_body,
			decorator_list = [],
			lineno     = self.__get_line(),
			col_offset = 0,
			)

		return func

	def __to_module(self):

		module_components = [
			ast.ImportFrom(module="ChromeController.transport",    names=[ast.alias('ChromeExecutionManager', None)], level=0),
			ast.ImportFrom(module="ChromeController.manager_base", names=[ast.alias('ChromeInterface',     None)], level=0),
			self.interface_class,
		]

		mod = ast.Module(module_components, lineno=self.__get_line(), col_offset=1)

		mod = ast.fix_missing_locations(mod)

		return mod

	def dump_class(self):
		indent = "	"
		return astor.to_source(self.__to_module(), indent_with=indent)

	def dump_ast(self):
		return astor.dump_tree(self.__to_module())

	def compile_class(self):
		mod = self.__to_module()
		code = compile(self.__to_module(), "no filename", "exec")
		exec(code)
		built_class = locals()['ChromeRemoteDebugInterface']

		return built_class

def get_source():
	instance = JsonInterfaceGenerator()
	return instance.dump_class()

def get_class_def():
	instance = JsonInterfaceGenerator()
	ret = instance.compile_class()
	return ret

def get_printed_ast():
	instance = JsonInterfaceGenerator()
	return instance.dump_ast()



def print_file_ast():
	with open(__file__) as fp:
		this_source = fp.read()
	this_ast = ast.parse(this_source)

	print("AST:")
	print("astor.dump_tree(this_ast)")
	print(astor.dump_tree(this_ast))

def update_generated_class(force=False, output_diff=False):
	log = logging.getLogger("Main.ChromeController.WrapperGenerator")
	gen_filename = "Generated.py"
	cur_file = os.path.abspath(__file__)
	cur_dir  = os.path.dirname(cur_file)

	fname = os.path.join(cur_dir, gen_filename)

	cls_def = get_source()
	try:
		with open(fname, "r", encoding='utf-8') as fp:
			have = fp.read()
	except IOError:
		# The class hasn't been generated yet?
		have = ""


	if have.strip() == cls_def.strip() and not force:
		log.info("ChromeController wrapper is up to date. Nothing to do")
	else:

		if not force:
			log.warning("Generated wrapper appears to be out of date. Regenerating.")
		else:
			log.warning("Generated wrapper being force-regenerated")
		log.warning("Note: If ChromeController is installed as a module, "
		                 "this may require administrator permissions")

		if output_diff:
			if have.strip() == cls_def.strip():
				log.info("No changes!")
			else:
				log.info("Calculating changes")
				d = difflib.Differ()
				diff = d.compare(have.strip().splitlines(keepends=True), cls_def.strip().splitlines(keepends=True))
				dstr = ''.join(diff)
				for line in dstr.split("\n"):
					if line.startswith(" "):
						continue
					log.info("Diff: %s", line)

		try:
			with open(fname, "w", encoding='utf-8') as fp:
				fp.write(cls_def)
		except IOError:
			if force:
				raise IOError("Could not update class definition file: {}, and "
				              "it is out of date!".format(fname))
			else:
				log.error("Failure updating file '%s'!" % fname)


def test():
	print(JsonInterfaceGenerator)
	# print_file_ast()
	instance = JsonInterfaceGenerator()
	print("ast:")
	print(instance.dump_ast())
	newsauce = instance.dump_class()
	print("Class:")
	print(newsauce)
	print(instance.compile_class())
	# print(instance)

if __name__ == '__main__':
	test()
	print_file_ast()







import logging
import contextlib
import traceback

from .manager import ChromeRemoteDebugInterface

@contextlib.contextmanager
def ChromeContext(*args, **kwargs):
	'''
	Context manager for conveniently handling the lifetime of the underlying chromium instance.

	In general, this should be the preferred way to use an instance of `ChromeRemoteDebugInterface`.

	All parameters are forwarded through to the underlying ChromeRemoteDebugInterface() constructor.
	'''
	log = logging.getLogger("Main.ChromeController.ChromeContext")
	chrome_created = False
	try:
		chrome_instance = ChromeRemoteDebugInterface(*args, **kwargs)
		chrome_created = True
		log.info("Entering chrome context")
		yield chrome_instance
	except Exception as e:

		log.error("Exception in chrome context!")
		for line in traceback.format_exc().split("\n"):
			log.error(line)
		raise e

	finally:
		log.info("Exiting chrome context")
		if chrome_created:
			chrome_instance.close()





class ChromeControllerException(RuntimeError):
	pass

class ChromeStartupException(ChromeControllerException):
	pass

class ReusedPortError(ChromeStartupException):
	pass

class ChromeConnectFailure(ChromeControllerException):
	pass

class ChromeCommunicationsError(ChromeControllerException):
	pass
class ChromeTabNotFoundError(ChromeCommunicationsError):
	pass

class ChromeError(ChromeControllerException):
	pass
class ChromeDiedError(ChromeError):
	pass

class ChromeNavigateTimedOut(ChromeError):
	pass
class ChromeResponseNotReceived(ChromeError):
	pass


"""
Utility (Linux only) to ensure subprocesses exit when their parents do by sending
a specified signal when the parent dies.

Usage:

    subprocess.Popen(['some-executable'], preexec_fn=on_parent_exit('SIGHUP'))

Taken from https://gist.github.com/evansd/2346614,
http://evans.io/legacy/posts/killing-child-processes-on-parent-exit-prctl/
"""

import signal
from ctypes import cdll

# Constant taken from http://linux.die.net/include/linux/prctl.h
PR_SET_PDEATHSIG = 1

class PrCtlError(Exception):
    pass

def on_parent_exit(signame):
    """
    Return a function to be run in a child process which will trigger SIGNAME
    to be sent when the parent process dies
    """
    signum = getattr(signal, signame)
    def set_parent_exit_signal():
        # http://linux.die.net/man/2/prctl
        result = cdll['libc.so.6'].prctl(PR_SET_PDEATHSIG, signum)
        if result != 0:
            raise PrCtlError('prctl failed with error code %s' % result)
    return set_parent_exit_signal




def capture_loading_events(message):
	if not message:
		return False
	if "method" not in message:
		return False
	if message['method'] != 'Network.requestWillBeSent':
		return False
	if 'params' not in message:
		return False
	mparams = message['params']
	if 'type' not in mparams:
		return False
	if 'requestId' not in mparams:
		return False
	if 'documentURL' not in mparams:
		return False
	if mparams['type'] == 'Document':
		# pprint.pprint(message['params'])
		return True
	return False



def check_frame_navigated_command(expected_id):
	def check_frame_navigated(message):
		if not message:
			return False
		if "method" not in message:
			return False
		if message['method'] != "Page.frameNavigated":
			return False
		if 'params' not in message:
			return False
		params = message['params']
		if 'frame' not in params:
			return False
		frame = params['frame']
		if 'id' in frame:
			ret = frame['id'] == expected_id
			# print('check_frame_navigated', message)
			return ret
		return False
	return check_frame_navigated

def check_frame_load_command(method_name):
	def frame_loading_tracker(message):
		if not message:
			return False
		if "method" not in message:
			return False
		if message['method'] != method_name:
			return False
		if 'params' not in message:
			return False
		return True

		# Disabled. See https://bugs.chromium.org/p/chromedriver/issues/detail?id=1387
		# params = message['params']
		# if 'frameId' not in params:
		# 	return False
		# if 'frameId' in params:
		# 	ret = params['frameId'] == expected_id
		# 	# print("frame_loading_tracker", message)
		# return False

	return frame_loading_tracker



def wait_for_methods(method_list):
	def wait_for_methods_tracker(message):
		if not message:
			return False
		if "method" not in message:
			return False
		if message['method'] not in method_list:
			return False
		if 'params' not in message:
			return False
		return True

	return wait_for_methods_tracker



def check_frame_loader_command(method_name, loader_id):
	def frame_loading_tracker_with_loader(message):
		if not message:
			return False
		if "method" not in message:
			return False
		if message['method'] != method_name:
			return False
		if 'params' not in message:
			return False
		if 'loaderId' not in message['params']:
			return False
		if message['params']["loaderId"] == loader_id:
			return True

		return False


	return frame_loading_tracker_with_loader

def check_load_event_fired(message):
	if not message:
		return False
	if "method" not in message:
		return False
	if message['method'] == 'Page.loadEventFired':
		# print("check_load_event_fired", message)
		return True
	return False


def network_response_recieved_for_url(url, expected_id):
	def network_response_recieved_tracker(message):
		if not message:
			return False
		if "method" not in message:
			return False
		if message['method'] != 'Network.responseReceived':
			return False
		if 'params' not in message:
			return False
		params = message['params']
		if 'frameId' not in params:
			return False
		if 'frameId' in params:
			if params['frameId'] == expected_id and 'response' in params:
				# Checking the url in the response breaks if
				# the remote issues a 301 or 302.
				if url:
					response = params['response']
					if 'url' in response:
						return url == response['url']
				else:
					return True

		return False
	return network_response_recieved_tracker



# // Copyright 2016 The Chromium Authors. All rights reserved.
# // Use of this source code is governed by a BSD-style license that can be
# // found in the LICENSE file.

# // This file was generated at (2016-11-18 07:04:40.019257) by running:
# //
# ../../chrome/test/chromedriver/embed_js_in_cpp.py --directory gen/chrome/test/chromedriver/chrome
# ../../chrome/test/chromedriver/js/add_cookie.js
# ../../chrome/test/chromedriver/js/call_function.js
# ../../chrome/test/chromedriver/js/execute_async_script.js
# ../../chrome/test/chromedriver/js/focus.js
# ../../chrome/test/chromedriver/js/get_element_region.js
# ../../chrome/test/chromedriver/js/is_option_element_toggleable.js

#include "chrome/test/chromedriver/chrome/js.h"

kFocusScript = '''
    function() { // Copyright (c) 2013 The Chromium Authors. All rights reserved.
    // Use of this source code is governed by a BSD-style license that can be
    // found in the LICENSE file.

    function focus(element) {
      // Focus the target element in order to send keys to it.
      // First, the currently active element is blurred, if it is different from
      // the target element. We do not want to blur an element unnecessarily,
      // because this may cause us to lose the current cursor position in the
      // element.
      // Secondly, we focus the target element.
      // Thirdly, if the target element is newly focused and is a text input, we
      // set the cursor position at the end.
      // Fourthly, we check if the new active element is the target element. If not,
      // we throw an error.
      // Additional notes:
      //   - |document.activeElement| is the currently focused element, or body if
      //     no element is focused
      //   - Even if |document.hasFocus()| returns true and the active element is
      //     the body, sometimes we still need to focus the body element for send
      //     keys to work. Not sure why
      //   - You cannot focus a descendant of a content editable node
      //   - V8 throws a TypeError when calling setSelectionRange for a non-text
      //     input, which still have setSelectionRange defined. For chrome 29+, V8
      //     throws a DOMException with code InvalidStateError.
      var doc = element.ownerDocument || element;
      var prevActiveElement = doc.activeElement;
      if (element != prevActiveElement && prevActiveElement)
        prevActiveElement.blur();
      element.focus();
      if (element != prevActiveElement && element.value &&
          element.value.length && element.setSelectionRange) {
        try {
          element.setSelectionRange(element.value.length, element.value.length);
        } catch (error) {
          if (!(error instanceof TypeError) && !(error instanceof DOMException &&
              error.code == DOMException.INVALID_STATE_ERR))
            throw error;
        }
      }

      var activeElement = doc.activeElement;
      // If the element is in a shadow DOM, then as far as the document is
      // concerned, the shadow host is the active element. We need to go through the
      // tree of shadow DOMs to check that the element we gave focus to is now
      // active.
      if (element != activeElement) {
        var shadowRoot = activeElement.shadowRoot;
        while (shadowRoot) {
          var activeElement = shadowRoot.activeElement;
          if (element == activeElement) {
            // the shadow DOM's active element is our requested element. We're good.
            break;
          }
          // The shadow DOM's active element isn't our requested element, check to
          // see if there's a nested shadow DOM.
          shadowRoot = activeElement.shadowRoot;
        }
      }
      if (element != activeElement)
        throw new Error('cannot focus element');
    }
    ; return focus.apply(null, arguments) }

    '''
kGetElementRegionScript = '''
    function() { // Copyright (c) 2013 The Chromium Authors. All rights reserved.
    // Use of this source code is governed by a BSD-style license that can be
    // found in the LICENSE file.

    function getElementRegion(element) {
      // Check that node type is element.
      if (element.nodeType != 1)
        throw new Error(element + ' is not an element');

      // We try 2 methods to determine element region. Try the first client rect,
      // and then the bounding client rect.
      // SVG is one case that doesn't have a first client rect.
      var clientRects = element.getClientRects();

      // Element area of a map has same first ClientRect and BoundingClientRect
      // after blink roll at chromium commit position 290738 which includes blink
      // revision 180610. Thus handle area as a special case.
      if (clientRects.length == 0 || element.tagName.toLowerCase() == 'area') {
        var box = element.getBoundingClientRect();
        if (element.tagName.toLowerCase() == 'area') {
          var coords = element.coords.split(',');
          if (element.shape.toLowerCase() == 'rect') {
            if (coords.length != 4)
              throw new Error('failed to detect the region of the area');
            var leftX = Number(coords[0]);
            var topY = Number(coords[1]);
            var rightX = Number(coords[2]);
            var bottomY = Number(coords[3]);
            return {
                'left': leftX,
                'top': topY,
                'width': rightX - leftX,
                'height': bottomY - topY
            };
          } else if (element.shape.toLowerCase() == 'circle') {
            if (coords.length != 3)
              throw new Error('failed to detect the region of the area');
            var centerX = Number(coords[0]);
            var centerY = Number(coords[1]);
            var radius = Number(coords[2]);
            return {
                'left': Math.max(0, centerX - radius),
                'top': Math.max(0, centerY - radius),
                'width': radius * 2,
                'height': radius * 2
            };
          } else if (element.shape.toLowerCase() == 'poly') {
            if (coords.length < 2)
              throw new Error('failed to detect the region of the area');
            var minX = Number(coords[0]);
            var minY = Number(coords[1]);
            var maxX = minX;
            var maxY = minY;
            for (i = 2; i < coords.length; i += 2) {
              var x = Number(coords[i]);
              var y = Number(coords[i + 1]);
              minX = Math.min(minX, x);
              minY = Math.min(minY, y);
              maxX = Math.max(maxX, x);
              maxY = Math.max(maxY, y);
            }
            return {
                'left': minX,
                'top': minY,
                'width': maxX - minX,
                'height': maxY - minY
            };
          } else {
            throw new Error('shape=' + element.shape + ' is not supported');
          }
        }
        return {
            'left': 0,
            'top': 0,
            'width': box.width,
            'height': box.height
        };
      } else {
        var box = element.getBoundingClientRect();
        var clientRect = clientRects[0];
        for (var i = 0; i < clientRects.length; i++) {
          if (clientRects[i].height != 0 && clientRects[i].width != 0) {
            clientRect = clientRects[i];
            break;
          }
        }
        return {
            'left': clientRect.left - box.left,
            'top': clientRect.top - box.top,
            'width': clientRect.right - clientRect.left,
            'height': clientRect.bottom - clientRect.top
        };
      }
    }
    ; return getElementRegion.apply(null, arguments) }

    '''
kIsOptionElementToggleableScript = '''
    function() { // Copyright (c) 2013 The Chromium Authors. All rights reserved.
    // Use of this source code is governed by a BSD-style license that can be
    // found in the LICENSE file.

    function isOptionElementToggleable(option) {
      if (option.tagName.toLowerCase() != 'option')
        throw new Error('element is not an option');
      for (var parent = option.parentElement;
           parent;
           parent = parent.parentElement) {
        if (parent.tagName.toLowerCase() == 'select') {
          return parent.multiple;
        }
      }
      throw new Error('option element is not in a select');
    }
    ; return isOptionElementToggleable.apply(null, arguments) }

    '''
kExecuteAsyncScriptScript = '''
    function() { // Copyright (c) 2013 The Chromium Authors. All rights reserved.
    // Use of this source code is governed by a BSD-style license that can be
    // found in the LICENSE file.

    /**
     * Enum for WebDriver status codes.
     * @enum {number}
     */
    var StatusCode = {
      OK: 0,
      UNKNOWN_ERROR: 13,
      JAVASCRIPT_ERROR: 17,
      SCRIPT_TIMEOUT: 28,
    };

    /**
     * Dictionary key for asynchronous script info.
     * @const
     */
    var ASYNC_INFO_KEY = '$chrome_asyncScriptInfo';

    /**
    * Return the information of asynchronous script execution.
    *
    * @return {Object<?>} Information of asynchronous script execution.
    */
    function getAsyncScriptInfo() {
      if (!(ASYNC_INFO_KEY in document))
        document[ASYNC_INFO_KEY] = {'id': 0};
      return document[ASYNC_INFO_KEY];
    }

    /**
    * Execute the given script and save its asynchronous result.
    *
    * If script1 finishes after script2 is executed, then script1's result will be
    * discarded while script2's will be saved.
    *
    * @param {string} script The asynchronous script to be executed. The script
    *     should be a proper function body. It will be wrapped in a function and
    *     invoked with the given arguments and, as the final argument, a callback
    *     function to invoke to report the asynchronous result.
    * @param {!Array<*>} args Arguments to be passed to the script.
    * @param {boolean} isUserSupplied Whether the script is supplied by the user.
    *     If not, UnknownError will be used instead of JavaScriptError if an
    *     exception occurs during the script, and an additional error callback will
    *     be supplied to the script.
    * @param {?number} opt_timeoutMillis The timeout, in milliseconds, to use.
    *     If the timeout is exceeded and the callback has not been invoked, a error
    *     result will be saved and future invocation of the callback will be
    *     ignored.
    */
    function executeAsyncScript(script, args, isUserSupplied, opt_timeoutMillis) {
      var info = getAsyncScriptInfo();
      info.id++;
      delete info.result;
      var id = info.id;

      function report(status, value) {
        if (id != info.id)
          return;
        info.id++;
        info.result = {status: status, value: value};
      }
      function reportValue(value) {
        report(StatusCode.OK, value);
      }
      function reportScriptError(error) {
        var code = isUserSupplied ? StatusCode.JAVASCRIPT_ERROR :
                                    (error.code || StatusCode.UNKNOWN_ERROR);
        var message = error.message;
        if (error.stack) {
          message += \"\\nJavaScript stack:\\n\" + error.stack;
        }
        report(code, message);
      }
      args.push(reportValue);
      if (!isUserSupplied)
        args.push(reportScriptError);

      try {
        new Function(script).apply(null, args);
      } catch (error) {
        reportScriptError(error);
        return;
      }

      if (typeof(opt_timeoutMillis) != 'undefined') {
        window.setTimeout(function() {
          var code = isUserSupplied ? StatusCode.SCRIPT_TIMEOUT :
                                      StatusCode.UNKNOWN_ERROR;
          var errorMsg = 'result was not received in ' + opt_timeoutMillis / 1000 +
                         ' seconds';
          report(code, errorMsg);
        }, opt_timeoutMillis);
      }
    }
    ; return executeAsyncScript.apply(null, arguments) }

    '''
kCallFunctionScript = '''
    function() { // Copyright (c) 2012 The Chromium Authors. All rights reserved.
    // Use of this source code is governed by a BSD-style license that can be
    // found in the LICENSE file.

    /**
     * Enum for WebDriver status codes.
     * @enum {number}
     */
    var StatusCode = {
      STALE_ELEMENT_REFERENCE: 10,
      UNKNOWN_ERROR: 13,
    };

    /**
     * Enum for node types.
     * @enum {number}
     */
    var NodeType = {
      ELEMENT: 1,
      DOCUMENT: 9,
    };

    /**
     * Dictionary key to use for holding an element ID.
     * @const
     * @type {string}
     */
    var ELEMENT_KEY = 'ELEMENT';

    /**
     * True if shadow dom is enabled.
     * @const
     * @type {boolean}
     */
    var SHADOW_DOM_ENABLED = typeof ShadowRoot === 'function';

    /**
     * A cache which maps IDs <-> cached objects for the purpose of identifying
     * a script object remotely.
     * @constructor
     */
    function Cache() {
      this.cache_ = {};
      this.nextId_ = 1;
      this.idPrefix_ = Math.random().toString();
    }

    Cache.prototype = {

      /**
       * Stores a given item in the cache and returns a unique ID.
       *
       * @param {!Object} item The item to store in the cache.
       * @return {number} The ID for the cached item.
       */
      storeItem: function(item) {
        for (var i in this.cache_) {
          if (item == this.cache_[i])
            return i;
        }
        var id = this.idPrefix_  + '-' + this.nextId_;
        this.cache_[id] = item;
        this.nextId_++;
        return id;
      },

      /**
       * Retrieves the cached object for the given ID.
       *
       * @param {number} id The ID for the cached item to retrieve.
       * @return {!Object} The retrieved item.
       */
      retrieveItem: function(id) {
        var item = this.cache_[id];
        if (item)
          return item;
        var error = new Error('not in cache');
        error.code = StatusCode.STALE_ELEMENT_REFERENCE;
        error.message = 'element is not attached to the page document';
        throw error;
      },

      /**
       * Clears stale items from the cache.
       */
      clearStale: function() {
        for (var id in this.cache_) {
          var node = this.cache_[id];
          if (!this.isNodeReachable_(node))
            delete this.cache_[id];
        }
      },

      /**
        * @private
        * @param {!Node} node The node to check.
        * @return {boolean} If the nodes is reachable.
        */
      isNodeReachable_: function(node) {
        var nodeRoot = getNodeRootThroughAnyShadows(node);
        return (nodeRoot == document);
      }
    };

    /**
     * Returns the root element of the node.  Found by traversing parentNodes until
     * a node with no parent is found.  This node is considered the root.
     * @param {!Node} node The node to find the root element for.
     * @return {!Node} The root node.
     */
    function getNodeRoot(node) {
      while (node.parentNode) {
        node = node.parentNode;
      }
      return node;
    }

    /**
     * Returns the root element of the node, jumping up through shadow roots if
     * any are found.
     */
    function getNodeRootThroughAnyShadows(node) {
      var root = getNodeRoot(node);
      while (SHADOW_DOM_ENABLED && root instanceof ShadowRoot) {
        root = getNodeRoot(root.host);
      }
      return root;
    }

    /**
     * Returns the global object cache for the page.
     * @param {Document=} opt_doc The document whose cache to retrieve. Defaults to
     *     the current document.
     * @return {!Cache} The page's object cache.
     */
    function getPageCache(opt_doc) {
      var doc = opt_doc || document;
      var key = '$cdc_asdjflasutopfhvcZLmcfl_';
      if (!(key in doc))
        doc[key] = new Cache();
      return doc[key];
    }

    /**
     * Wraps the given value to be transmitted remotely by converting
     * appropriate objects to cached object IDs.
     *
     * @param {*} value The value to wrap.
     * @return {*} The wrapped value.
     */
    function wrap(value) {
      // As of crrev.com/1316933002, typeof() for some elements will return
      // 'function', not 'object'. So we need to check for both non-null objects, as
      // well Elements that also happen to be callable functions (e.g. <embed> and
      // <object> elements). Note that we can not use |value instanceof Object| here
      // since this does not work with frames/iframes, for example
      // frames[0].document.body instanceof Object == false even though
      // typeof(frames[0].document.body) == 'object'.
      if ((typeof(value) == 'object' && value != null) ||
          (typeof(value) == 'function' && value instanceof Element)) {
        var nodeType = value['nodeType'];
        if (nodeType == NodeType.ELEMENT || nodeType == NodeType.DOCUMENT
            || (SHADOW_DOM_ENABLED && value instanceof ShadowRoot)) {
          var wrapped = {};
          var root = getNodeRootThroughAnyShadows(value);
          wrapped[ELEMENT_KEY] = getPageCache(root).storeItem(value);
          return wrapped;
        }

        var obj = (typeof(value.length) == 'number') ? [] : {};
        for (var prop in value)
          obj[prop] = wrap(value[prop]);
        return obj;
      }
      return value;
    }

    /**
     * Unwraps the given value by converting from object IDs to the cached
     * objects.
     *
     * @param {*} value The value to unwrap.
     * @param {Cache} cache The cache to retrieve wrapped elements from.
     * @return {*} The unwrapped value.
     */
    function unwrap(value, cache) {
      if (typeof(value) == 'object' && value != null) {
        if (ELEMENT_KEY in value)
          return cache.retrieveItem(value[ELEMENT_KEY]);

        var obj = (typeof(value.length) == 'number') ? [] : {};
        for (var prop in value)
          obj[prop] = unwrap(value[prop], cache);
        return obj;
      }
      return value;
    }

    /**
     * Calls a given function and returns its value.
     *
     * The inputs to and outputs of the function will be unwrapped and wrapped
     * respectively, unless otherwise specified. This wrapping involves converting
     * between cached object reference IDs and actual JS objects. The cache will
     * automatically be pruned each call to remove stale references.
     *
     * @param  {Array<string>} shadowHostIds The host ids of the nested shadow
     *     DOMs the function should be executed in the context of.
     * @param {function(...[*]) : *} func The function to invoke.
     * @param {!Array<*>} args The array of arguments to supply to the function,
     *     which will be unwrapped before invoking the function.
     * @param {boolean=} opt_unwrappedReturn Whether the function's return value
     *     should be left unwrapped.
     * @return {*} An object containing a status and value property, where status
     *     is a WebDriver status code and value is the wrapped value. If an
     *     unwrapped return was specified, this will be the function's pure return
     *     value.
     */
    function callFunction(shadowHostIds, func, args, opt_unwrappedReturn) {
      var cache = getPageCache();
      cache.clearStale();
      if (shadowHostIds && SHADOW_DOM_ENABLED) {
        for (var i = 0; i < shadowHostIds.length; i++) {
          var host = cache.retrieveItem(shadowHostIds[i]);
          // TODO(zachconrad): Use the olderShadowRoot API when available to check
          // all of the shadow roots.
          cache = getPageCache(host.webkitShadowRoot);
          cache.clearStale();
        }
      }

      if (opt_unwrappedReturn)
        return func.apply(null, unwrap(args, cache));

      var status = 0;
      try {
        var returnValue = wrap(func.apply(null, unwrap(args, cache)));
      } catch (error) {
        status = error.code || StatusCode.UNKNOWN_ERROR;
        var returnValue = error.message;
      }
      return {
          status: status,
          value: returnValue
      }
    }
    ; return callFunction.apply(null, arguments) }

    '''
kAddCookieScript = '''
    function() { // Copyright (c) 2013 The Chromium Authors. All rights reserved.
    // Use of this source code is governed by a BSD-style license that can be
    // found in the LICENSE file.

    /**
    * Test whether the given domain is valid for a cookie.
    *
    * @param {string} domain Domain for a cookie.
    * @return {boolean} True if the domain is valid, otherwise false.
    */
    function isDomainValid(domain) {
      var dummyCookie = 'ChromeDriverwjers908fljsdf37459fsdfgdfwru=';

      document.cookie = dummyCookie + '; domain=' + domain;
      if (document.cookie.indexOf(dummyCookie) != -1) {
        // Expire the dummy cookie if it is added successfully.
        document.cookie = dummyCookie + '; Max-Age=0; domain=' + domain;
        return true;
      }
      return false;
    }

    /**
    * Add the given cookie to the current web page.
    *
    * If path is not specified, default to '/'.
    * If domain is not specified, default to document.domain, otherwise remove its
    * port number.
    *
    * Validate name, value, domain and path of the cookie in the same way as the
    * method CanonicalCookie::Create in src/net/cookies/canonical_cookie.cc. Besides
    * the following requirements, name, value, domain and path of the cookie should
    * not start or end with ' ' or '\\t', and should not contain '\\n', '\\r', or '\\0'.
    * <ul>
    * <li>name: no ';' or '='
    * <li>value: no ';'
    * <li>path: starts with '/', no ';'
    * </ul>
    *
    * @param {!Object} cookie An object representing a Cookie JSON Object as
    *     specified in https://code.google.com/p/selenium/wiki/JsonWireProtocol.
    */
    function addCookie(cookie) {
      function isNameInvalid(value) {
        return /(^[ \\t])|([;=\\n\\r\\0])|([ \\t]$)/.test(value);
      }
      function isValueInvalid(value) {
        return /(^[ \\t])|([;\\n\\r\\0])|([ \\t]$)/.test(value);
      }
      function isPathInvalid(path) {
        return path[0] != '/' || /([;\\n\\r\\0])|([ \\t]$)/.test(path);
      }

      var name = cookie['name'];
      if (!name || isNameInvalid(name))
        throw new Error('name of cookie is missing or invalid:\"' + name + '\"');

      var value = cookie['value'] || '';
      if (isValueInvalid(value))
        throw new Error('value of cookie is invalid:\"' + value + '\"');

      var domain = cookie['domain'];
      // Remove the port number from domain.
      if (domain) {
        var domain_parts = domain.split(':');
        if (domain_parts.length > 2)
          throw new Error('domain of cookie has too many colons');
        else if (domain_parts.length == 2)
          domain = domain_parts[0];
      }
      // Validate domain.
      if (domain && (isValueInvalid(domain) || !isDomainValid(domain))) {
        var error = new Error();
        error.code = 24;  // Error code for InvalidCookieDomain.
        error.message = 'invalid domain:\"' + domain + '\"';
        throw error;
      }

      var path = cookie['path'];
      if (path && isPathInvalid(path))
        throw new Error('path of cookie is invalid:\"' + path + '\"');

      var newCookie = name + '=' + value;
      newCookie += '; path=' + (path || '/');
      if (domain)
        newCookie += '; domain=' + domain;
      if (cookie['expiry']) {
        var expiredDate = new Date(cookie['expiry'] * 1000);
        newCookie += '; expires=' + expiredDate.toUTCString();
      }
      if (cookie['secure'])
        newCookie += '; secure';

      document.cookie = newCookie;
    }
    ; return addCookie.apply(null, arguments) }
    '''
